import os
import json
import pytz
import typing
import strax
import fsspec
import numbers
import straxen
import inspect
import tarfile
import tempfile
import warnings

import numpy as np
import pandas as pd

from urllib.parse import urlparse, parse_qs
from immutabledict import immutabledict

from ast import literal_eval
from strax.config import OMITTED
from utilix import xent_collection
from scipy.interpolate import interp1d
from straxen.misc import filter_kwargs
from typing import Container, Mapping, Union, Iterable

from pydantic.validators import find_validators
from pydantic.config import get_config
from pydantic.errors import PydanticTypeError

export, __all__ = strax.exporter()

_CACHES = {}

WARN = True


@export
def clear_config_caches():
    for cache in _CACHES.values():
        cache.clear()


@export
def config_cache_size_mb():
    return straxen.total_size(_CACHES) // 1e6


def parse_val(val: str):
    """Attempt to parse a string value as
    a python literal, falls back to returning just
    the original string if cant be parsed.
    """
    try:
        val = literal_eval(val)
    except ValueError:
        pass
    except SyntaxError:
        pass
    return val


def get_item_or_attr(obj, key, default=None):
    if isinstance(obj, dict):
        return obj.get(key, default)
    return getattr(obj, key, default)


@export
class URLConfig(strax.Config):
    """Dispatch on URL protocol.
    unrecognized protocol returns identity
    inspired by dasks Dispatch and fsspec fs protocols.
    """

    _LOOKUP = {}
    _PREPROCESSORS = ()

    SCHEME_SEP = "://"
    QUERY_SEP = "?"
    NAMESPACE_SEP = "."
    PLUGIN_ATTR_PREFIX = "plugin."

    def __init__(self, cache=0, **kwargs):
        """
        :param cache: number of values to keep in cache,
                      if set to True will cache all values
        :param **kwargs: additional keyword arguments accepted by strax.Option
        """
        self.final_type = OMITTED
        super().__init__(**kwargs)
        # Ensure backwards compatibility with Option validation
        # type of the config value can be different from the fetched value.
        if self.type is not OMITTED:
            self.final_type = self.type
            self.type = OMITTED  # do not enforce type on the URL
        if cache:
            cache_len = 100 if cache is True else int(cache)
            cache = straxen.CacheDict(cache_len=cache_len)
            _CACHES[id(self)] = cache

    @property
    def cache(self):
        return _CACHES.get(id(self), {})

    @classmethod
    def register(cls, protocol, func=None):
        """Register dispatch of `func` on urls
        starting with protocol name `protocol`"""

        def wrapper(func):
            if isinstance(protocol, tuple):
                for t in protocol:
                    cls.register(t, func)
                return func

            if not isinstance(protocol, str):
                raise ValueError("Protocol name must be a string.")

            if protocol in cls._LOOKUP:
                raise ValueError(f"Protocol with name {protocol} already registered.")
            cls._LOOKUP[protocol] = func
            return func

        return wrapper(func) if func is not None else wrapper

    @classmethod
    def preprocessor(cls, func=None, precedence=0):
        """Register a new processor to modify the config values
        before they are used.
        """

        def wrapper(func):
            entry = (precedence, func)
            if entry in cls._PREPROCESSORS:
                raise ValueError(f"This processor is already registered.")
            cls._PREPROCESSORS += (entry,)
            return func

        return wrapper(func) if func is not None else wrapper

    @classmethod
    def eval(cls, protocol: str, arg: Union[str, tuple] = None, kwargs: dict = None):
        """Evaluate a URL/AST by recusively dispatching protocols by name
            with argument arg and keyword arguments kwargs
           and return the value. If protocol does not exist, returnes arg
        :param protocol: name of the protocol or a URL
        :param arg: argument to pass to protocol, can be another (sub-protocol,
            arg, kwargs) tuple, in which case sub-protocol will be evaluated
            and passed to protocol
        :param kwargs: keyword arguments to be passed to the protocol
        :return: (Any) The return value of the protocol on these arguments
        """

        if protocol is not None and arg is None:
            protocol, arg, kwargs = cls.url_to_ast(protocol)

        if protocol is None:
            return arg

        if kwargs is None:
            kwargs = {}

        meth = cls._LOOKUP[protocol]

        if isinstance(arg, tuple):
            arg = cls.eval(*arg)

        # Just to be on the safe side
        kwargs = straxen.filter_kwargs(meth, kwargs)

        return meth(arg, **kwargs)

    @classmethod
    def split_url_kwargs(cls, url):
        """split a url into path and kwargs"""
        path, _, _ = url.partition(cls.QUERY_SEP)
        kwargs = {}
        for k, v in parse_qs(urlparse(url).query).items():
            # values of query arguments are evaluated as lists
            # split logic depending on length
            n = len(v)
            if not n:
                kwargs[k] = None
            elif n == 1:
                kwargs[k] = parse_val(v[0])
            else:
                kwargs[k] = list(map(parse_val, v))
        return path, kwargs

    @classmethod
    def kwarg_from_url(cls, url: str, key: str):
        path, kwargs = cls.split_url_kwargs(url)
        return kwargs.get(key, None)

    @classmethod
    def format_url_kwargs(cls, url, **kwargs):
        """Add keyword arguments to a URL.
        Sorts all arguments by key for hash consistency
        """
        url, extra_kwargs = cls.split_url_kwargs(url)
        kwargs = dict(extra_kwargs, **kwargs)
        arg_list = []
        for k, v in sorted(kwargs.items()):
            if isinstance(v, list):
                # lists are passed as multiple arguments with the same key
                arg_list.extend([f"{k}={vi}" for vi in v])
            else:
                arg_list.append(f"{k}={v}")
        arg_str = "&".join(arg_list)
        arg_str = cls.QUERY_SEP + arg_str if arg_str else ""
        return url + arg_str

    @classmethod
    def lookup_value(cls, value, **namespace):
        """Optionally fetch an attribute from namespace
        if value is a string with cls.NAMESPACE_SEP in
        it, the string is split and the first part is used
        to lookup an object in namespace and the second part
        is used to lookup the value in the object.
        If the value is not a string or the target object is
        not in the namesapce, the value is returned as is.
        """

        if isinstance(value, list):
            return [cls.lookup_value(v, **namespace) for v in value]

        if isinstance(value, str) and cls.NAMESPACE_SEP in value:
            name, _, key = value.partition(cls.NAMESPACE_SEP)
            if name in namespace:
                obj = namespace[name]
                if isinstance(obj, Mapping):
                    value = obj.get(key, value)
                else:
                    value = getattr(obj, key, value)

        return value

    @classmethod
    def deref_ast(cls, protocol, arg, kwargs, **namespace):
        """Dereference an AST by looking up values in namespace"""
        if isinstance(arg, tuple):
            arg = cls.deref_ast(*arg, **namespace)
        else:
            arg = cls.lookup_value(arg, **namespace)
        kwargs = {k: cls.lookup_value(v, **namespace) for k, v in kwargs.items()}
        return protocol, arg, kwargs

    def validate(
            self,
            config,
            run_id=None,  # TODO: will soon be removed
            run_defaults=None,
            set_defaults=True,
    ):
        """This method is called by the context on plugin initialization
        at this stage, the run_id and context config are already known but the
        config values are not yet set on the plugin. Therefore its the perfect
        place to run any preprocessors on the config values to make any needed
        changes before the configs are hashed.
        """
        super().validate(config, run_id, run_defaults, set_defaults)

        cfg = config[self.name]

        sorted_preprocessors = reversed(sorted(self._PREPROCESSORS, key=lambda x: x[0]))

        full_kwargs = dict(
            name=self.name,
            run_id=run_id,
            run_defaults=run_defaults,
            set_defaults=set_defaults,
        )

        for _, preprocessor in sorted_preprocessors:
            kwargs = filter_kwargs(preprocessor, full_kwargs)
            new_cfg = preprocessor(cfg, **kwargs)
            cfg = new_cfg if new_cfg is not None else cfg

        config[self.name] = cfg

        if not isinstance(cfg, str) or self.SCHEME_SEP not in cfg:
            # if the value is not a url config it is validated against
            # its intended type (final_type)
            self.validate_type(cfg)

    def validate_type(self, value):
        """Validate the type of a value against its intended type"""
        msg = (
            f"Invalid type for option {self.name}. "
            f"Expected a {self.final_type} instance, got {type(value)}"
        )

        if self.final_type is not OMITTED:
            # Use pydantic to validate the type
            # its validation is more flexible than isinstance
            # it will coerce standard equivalent types
            cfg = get_config(dict(arbitrary_types_allowed=True))
            if isinstance(self.final_type, tuple):
                validators = [v for t in self.final_type for v in find_validators(t, config=cfg)]
            else:
                validators = find_validators(self.final_type, config=cfg)
            for validator in validators:
                try:
                    validator(value)
                    break
                except:
                    pass
            else:
                raise TypeError(msg)

        return value

    def fetch(self, plugin):
        """override the Config.fetch method
        this is called when the attribute is accessed
        from withing the Plugin instance
        """
        # first fetch the user-set value

        # from the config dictionary
        url = super().fetch(plugin)

        if not isinstance(url, str):
            # if the value is not a string it is evaluated
            # as a literal config and returned as is.
            return self.validate_type(url)

        if self.SCHEME_SEP not in url:
            # no protocol in the url so its evaluated
            # as string-literal config and returned as is
            return self.validate_type(url)

        # evaluate the url as AST
        protocol, arg, kwargs = self.url_to_ast(url)

        # allow run_id to be missing
        run_id = getattr(plugin, "run_id", "000000")

        # construct a deterministic hash key from AST
        key = strax.deterministic_hash(
            (plugin.config, run_id, protocol, arg, kwargs)
        )

        # fetch from cache if exists
        value = self.cache.get(key, None)

        # not in cache, lets fetch it
        if value is None:
            # resolve any referenced to plugin or config attributes
            protocol, arg, kwargs = self.deref_ast(
                protocol, arg, kwargs, config=plugin.config, plugin=plugin
            )

            value = self.eval(protocol, arg, kwargs)
            value = self.validate_type(value)
            self.cache[key] = value

        return value

    @classmethod
    def ast_to_url(
            cls,
            protocol: Union[str, tuple],
            arg: Union[str, tuple] = None,
            kwargs: dict = None,
    ):
        """Convert a protocol abstract syntax tree to a valid URL"""

        if isinstance(protocol, tuple):
            protocol, arg, kwargs = protocol

        if kwargs is None:
            kwargs = {}

        if protocol is None:
            return arg

        if isinstance(arg, (list, dict, numbers.Number)) and protocol != "json":
            arg = (
                "json",
                json.dumps(arg),
            )

        if isinstance(arg, tuple):
            arg = cls.ast_to_url(*arg)

        if not isinstance(arg, str):
            raise TypeError(f"Type {type(arg)} is not supported as an argument.")

        arg, extra_kwargs = cls.split_url_kwargs(arg)

        kwargs.update(extra_kwargs)

        url = f"{protocol}{cls.SCHEME_SEP}{arg}"

        url = cls.format_url_kwargs(url, **kwargs)

        return url

    @classmethod
    def url_to_ast(cls, url, **kwargs):
        """Convert a URL to a protocol abstract syntax tree"""
        if not isinstance(url, str):
            raise TypeError(f"URL must be a string, got {type(url)}")

        if cls.SCHEME_SEP not in url:
            # no protocol in the url so its evaluated
            # as string-literal config and returned as is
            return None, url, {}

        # separate the protocol name from the path
        protocol, _, path = url.partition(cls.SCHEME_SEP)

        # find the corresponding protocol method
        meth = cls._LOOKUP.get(protocol, None)
        if meth is None:
            # unrecognized protocol
            # evaluate as string-literal
            return None, url, {}

        arg, url_kwargs = cls.split_url_kwargs(path)
        kwargs.update(url_kwargs)

        if cls.SCHEME_SEP in arg:
            # url contains a nested protocol
            # first parsce sub-protocol
            arg = cls.url_to_ast(arg, **kwargs)

        # Filter unused kwargs for this method.
        # This is done also at the eval level but
        # probably better to be safe.
        kwargs = straxen.filter_kwargs(meth, kwargs)

        # Always sort kwargs for consistent ASTs
        kwargs = dict(sorted(kwargs.items()))

        return protocol, arg, kwargs

    @classmethod
    def are_equal(cls, first, second):
        """Return whether two URLs are equivalent (have equal ASTs)"""
        return cls.url_to_ast(first) == cls.url_to_ast(second)

    @classmethod
    def protocol_descr(cls):
        rows = []
        for k, v in cls._LOOKUP.items():
            descr = v.__doc__
            if descr is not None:
                descr = descr.split("\n")[0]

            row = {
                "name": f"{k}://",
                "description": descr,
                "signature": str(inspect.signature(v)),
                "location": v.__module__,
            }
            rows.append(row)
        return pd.DataFrame(rows)

    @classmethod
    def print_protocols(cls):
        df = cls.protocol_descr()
        if len(df):
            print(df)
        else:
            print("No protocols registered.")

    @classmethod
    def preprocessor_descr(cls):
        rows = []
        for k, v in cls._PREPROCESSORS:
            descr = v.__doc__
            if descr is not None:
                descr = descr.split("\n")[0]
            row = {
                "precedence": k,
                "description": descr,
                "signature": str(inspect.signature(v)),
                "location": v.__module__,
            }
            rows.append(row)
        return pd.DataFrame(rows)

    @classmethod
    def print_preprocessors(cls):
        df = cls.preprocessor_descr()
        if len(df):
            print(df)
        else:
            print("No Preprocessors registered.")

    @classmethod
    def print_summary(cls):
        print("=" * 30 + " Protocols " + "=" * 30)
        cls.print_protocols()
        print("=" * 30 + " Preprocessors " + "=" * 30)
        cls.print_preprocessors()

    @classmethod
    def evaluate_dry(cls, url: str, **kwargs):
        """
        Utility function to quickly test and evaluate URL configs,
        without the initialization of plugins (so no plugin attributes).
        plugin attributes can be passed as keyword arguments.

        example::

            from straxen import URLConfig
            url_string='cmt://electron_drift_velocity?run_id=027000&version=v3'
            URLConfig.evaluate_dry(url_string)

            # or similarly
            url_string='cmt://electron_drift_velocity?run_id=plugin.run_id&version=v3'
            URLConfig.evaluate_dry(url_string, run_id='027000')

        Please note that this has to be done outside of the plugin, so any
        attributes of the plugin are not yet note to this dry evaluation
        of the url-string.

        :param url: URL to evaluate, see above for example.
        :keyword: any additional kwargs are passed to self.dispatch (see example)
        :return: evaluated value of the URL.
        """
        url = cls.format_url_kwargs(url, **kwargs)
        _, combined_kwargs = cls.split_url_kwargs(url)

        for k, v in combined_kwargs.items():
            if isinstance(v, str) and cls.PLUGIN_ATTR_PREFIX in v:
                raise ValueError(
                    f"The URL parameter {k} depends on the plugin. "
                    "You must specify the value for this parameter "
                    "for this URL to be evaluated correctly. "
                    f"Try passing {k} as a keyword argument. "
                    f"e.g.: `URLConfig.evaluate_dry({url}, {k}=SOME_VALUE)`."
                )

        return cls.eval(url)


@URLConfig.register("cmt")
def get_correction(
        name: str,
        run_id: str = None,
        version: str = "ONLINE",
        detector: str = "nt",
        **kwargs,
):
    """Get value for name from CMT"""

    if run_id is None:
        raise ValueError("Attempting to fetch a correction without a run id.")

    return straxen.get_correction_from_cmt(run_id, (name, version, detector == "nt"))


@URLConfig.register("resource")
def get_resource(name: str, fmt: str = "text", **kwargs):
    """
    Fetch a straxen resource
    Allow a direct download using <fmt='abs_path'> otherwise kwargs are
    passed directly to straxen.get_resource.
    """
    if fmt == "abs_path":
        downloader = straxen.MongoDownloader()
        return downloader.download_single(name)
    return straxen.get_resource(name, fmt=fmt)


@URLConfig.register("fsspec")
def read_file(path: str, **kwargs):
    """Support fetching files from arbitrary filesystems"""
    with fsspec.open(path, **kwargs) as f:
        content = f.read()
    return content


@URLConfig.register("json")
def read_json(content: str, **kwargs):
    """Load json string as a python object"""
    return json.loads(content)


@URLConfig.register("take")
def get_key(container: Container, take=None, **kwargs):
    """return a single element of a container"""
    if take is None:
        return container
    if not isinstance(take, list):
        take = [take]

    # support for multiple keys for
    # nested objects
    for t in take:
        container = container[t]

    return container


@URLConfig.register("format")
def format_arg(arg: str, **kwargs):
    """apply pythons builtin format function to a string"""
    return arg.format(**kwargs)


@URLConfig.register("itp_map")
def load_map(
        some_map, method="WeightedNearestNeighbors", scale_coordinates=None, **kwargs
):
    """Make an InterpolatingMap"""
    itp_map = straxen.InterpolatingMap(some_map, method=method, **kwargs)
    if scale_coordinates is not None:
        itp_map.scale_coordinates(scale_coordinates)
    return itp_map


@URLConfig.register("bodega")
def load_value(name: str, bodega_version=None):
    """Load a number from BODEGA file"""
    if bodega_version is None:
        raise ValueError("Provide version see e.g. tests/test_url_config.py")
    nt_numbers = straxen.get_resource("XENONnT_numbers.json", fmt="json")
    return nt_numbers[name][bodega_version]["value"]


@URLConfig.register("tf")
def open_neural_net(model_path: str, custom_objects=None, **kwargs):
    """Open a tensorflow file and return a keras model."""
    # Nested import to reduce loading time of import straxen and it not
    # base requirement
    import tensorflow as tf

    if not os.path.exists(model_path):
        raise FileNotFoundError(f"No file at {model_path}")
    with tempfile.TemporaryDirectory() as tmpdirname:
        tar = tarfile.open(model_path, mode="r:gz")
        tar.extractall(path=tmpdirname)
        return tf.keras.models.load_model(tmpdirname, custom_objects=custom_objects)


@URLConfig.register("itp_dict")
def get_itp_dict(
        loaded_json, run_id=None, time_key="time", itp_keys="correction", **kwargs
) -> typing.Union[np.ndarray, typing.Dict[str, np.ndarray]]:
    """
    Interpolate a dictionary at the start time that is queried from
    a run-id.

    :param loaded_json: a dictionary with a time-series
    :param run_id: run_id
    :param time_key: key that gives the timestamps
    :param itp_keys: which keys from the dict to read. Should be
        comma (',') separated!

    :return: Interpolated values of dict at the start time, either
        returned as an np.ndarray (single value) or as a dict
        (multiple itp_dict_keys)
    """
    keys = strax.to_str_tuple(itp_keys.split(","))
    for key in list(keys) + [time_key]:
        if key not in loaded_json:
            raise KeyError(
                f"The json does contain the key '{key}'. Try one of: {loaded_json.keys()}"
            )

    times = loaded_json[time_key]

    # get start time of this run. Need to make tz-aware
    start = xent_collection().find_one({"number": int(run_id)}, {"start": 1})["start"]
    start = pytz.utc.localize(start).timestamp() * 1e9

    try:
        if len(strax.to_str_tuple(keys)) > 1:
            return {
                key: interp1d(times, loaded_json[key], bounds_error=True)(start)
                for key in keys
            }

        else:
            interp = interp1d(times, loaded_json[keys[0]], bounds_error=True)
            return interp(start)
    except ValueError as e:
        raise ValueError(f"Correction is not defined for run {run_id}") from e


@URLConfig.register("rekey_dict")
def rekey_dict(d, replace_keys="", with_keys=""):
    """Replace the keys of a dictionary.
    :param d: dictionary that will have its keys renamed
    :param replace_keys: comma-separated string of keys that will be replaced
    :param with_keys:  comma-separated string of keys that will replace the replace_keys
    :return: dictionary with renamed keys
    """
    new_dict = d.copy()
    replace_keys = strax.to_str_tuple(replace_keys.split(","))
    with_keys = strax.to_str_tuple(with_keys.split(","))
    if len(replace_keys) != len(with_keys):
        raise RuntimeError("replace_keys and with_keys must have the same length")
    for old_key, new_key in zip(replace_keys, with_keys):
        new_dict[new_key] = new_dict.pop(old_key)
    return new_dict


@URLConfig.register("objects-to-dict")
def objects_to_dict(objects: list, key_attr=None, value_attr="value", immutable=False):
    """Converts a list of objects/dicts to a single dictionary by taking the
    key and value from each of the objects/dicts. If key_attr is not provided,
    the list index is used as the key.

    :param objects: list of objects/dicts that will be converted to a dictionary
    :param key_attr: key/attribute of the objects that will be used as key in the dictionary
    :param value_attr: key/attribute of the objects that will be used as value in the dictionary
    """
    if not isinstance(objects, Iterable):
        raise TypeError(
            f"The objects-to-dict protocol expects an iterable "
            f"of objects but received {type(objects)} instead."
        )
    result = {}
    for i, obj in enumerate(objects):
        key = i if key_attr is None else get_item_or_attr(obj, key_attr)
        result[key] = get_item_or_attr(obj, value_attr)

    if immutable:
        result = immutabledict(result)

    return result


@URLConfig.register("list-to-array")
def objects_to_array(objects: list):
    """
    Converts a list of objects/dicts to a numpy array.
    :param objects: Any list of objects"""

    if not isinstance(objects, Iterable):
        raise TypeError(
            f"The list-to-array protocol expects an "
            f"iterable but recieved a {type(objects)} instead"
        )

    return np.array(objects)


@URLConfig.preprocessor
def alphabetize_url_kwargs(url: str):
    """
    Reorders queries for urlconfigs to avoid hashing issues
    """

    global WARN

    if isinstance(url, str) and URLConfig.SCHEME_SEP in url:
        if url != URLConfig.format_url_kwargs(url) and WARN:
            warnings.warn(
                "From straxen version 2.1.0 onward, URLConfig parameters"
                "will be sorted alphabetically before being passed to the plugins,"
                " this will change the lineage hash for non-sorted URLs. To load"
                " data processed with non-sorted URLs, you will need to use an"
                " older version."
            )
            WARN = False
        return URLConfig.format_url_kwargs(url)
    return url


@URLConfig.register("run_doc")
def read_rundoc(path, run_id=None, default=None):
    """Read a path from the rundoc."""
    if run_id is None:
        raise ValueError("rundoc protocol: missing run_id.")
    runs = xent_collection()
    rundoc = runs.find_one({"number": int(run_id)}, {"_id": 0, path: 1})
    if rundoc is None:
        raise ValueError(f"No rundoc found for run {run_id}")

    for part in path.split("."):
        if isinstance(rundoc, list) and part.isdigit() and len(rundoc) > int(part):
            rundoc = rundoc[int(part)]
        elif isinstance(rundoc, dict) and part in rundoc:
            rundoc = rundoc[part]
        elif default is not None:
            return default
        else:
            raise ValueError(f"No path {path} found in rundoc for run {run_id}")
    return rundoc


@URLConfig.register('pad-array')
def get_paded_array(arr: np.ndarray, pad_value=0, pad_left=0, pad_right=0):
    """Pad the array with pad_value on the left and right side"""
    return np.pad(arr, (pad_left, pad_right), constant_values=pad_value)



"""S1 Position reconstruction for Xenon-nT
Author and maintainer: Matteo Guida (guidam@mpi-hd.mpg.de)"""

import strax
import straxen
import numpy as np
from warnings import warn
export, __all__ = strax.exporter()

@export
class EventS1PositionBase(strax.Plugin):
    """
    Base pluging for S1 position reconstruction at event level
    """
    __version__ = '0.0.0'
    depends_on = ('event_area_per_channel', 'event_basics')

    algorithm = None
    compressor = 'zstd'
    parallel = True  # can set to "process" after #82

    min_s1_area_s1_posrec = straxen.URLConfig(
        help='Skip reconstruction if area (PE) is less than this',
        default=1000, infer_type=False, )
    n_top_pmts = straxen.URLConfig(
        default=straxen.n_top_pmts, infer_type=False,
        help="Number of top PMTs")

    def infer_dtype(self):
        if self.algorithm is None:
            raise NotImplementedError(f'Base class should not be used without '
                                      f'algorithm as done in {__class__.__name__}')

    def infer_dtype(self):
        if self.algorithm is None:
            raise NotImplementedError(f'Base class should not be used without '
                                      f'algorithm as done in {__class__.__name__}')
        dtype = [('event_x_' + self.algorithm, np.float32,
                  f'Reconstructed {self.algorithm} S1 X position (cm), uncorrected'),
                 ('event_y_' + self.algorithm, np.float32,
                  f'Reconstructed {self.algorithm} S1 Y position (cm), uncorrected'),
                 ('event_z_' + self.algorithm, np.float32,
                  f'Reconstructed {self.algorithm} S1 Z position (cm), uncorrected')
                  ]

        dtype += strax.time_fields
        return dtype

    def get_tf_model(self):
        """
        Simple wrapper to have several tf_event_model_s1_cnn, ..
        point to this same function in the compute method
        """
        model = getattr(self, f'tf_event_model_{self.algorithm}', None)
        if model is None:
            warn(f'Setting model to None for {self.__class__.__name__} will '
                 f'set only nans as output for {self.algorithm}')
        if isinstance(model, str):
            raise ValueError(f'open files from tf:// protocol! Got {model} '
                             f'instead, see tests/test_s1_posrec.py for examples.')
        return model

    def compute(self,events):
        result = np.ones(len(events), dtype=self.dtype)
        result['time'], result['endtime'] = events['time'], strax.endtime(events)

        result['event_x_' + self.algorithm] *= float('nan')
        result['event_y_' + self.algorithm] *= float('nan')
        result['event_z_' + self.algorithm] *= float('nan')

        model = self.get_tf_model()

        # Reconstruct position only for large peaks, otherwise severe inaccuracy.
        event_mask = events['s1_area_per_channel'].sum(axis=1) > self.min_s1_area_s1_posrec

        if not np.sum(event_mask):
            # No peaks fulfilling the conditions, return nan array.
            return result

        _in = events['s1_area_per_channel'][event_mask]

        with np.errstate(divide='ignore', invalid='ignore'):
        # Normalise patters by dividing by largest PMT output between the two arrays. 
            _in = _in / _in.max(axis=1,keepdims=True)

        # Getting actual position reconstruction
        _out = model.predict(_in)

        # writing output to the result
        result['event_x_' + self.algorithm][event_mask] = _out[:, 0]
        result['event_y_' + self.algorithm][event_mask] = _out[:, 1]
        result['event_z_' + self.algorithm][event_mask] = _out[:, 2]    
        return result



import numba
import numpy as np
import strax
from immutabledict import immutabledict
from strax.processing.general import _touching_windows
import straxen
from straxen.plugins.defaults import HE_PREAMBLE
from straxen.plugins.peaks.peaks import Peaks
from straxen.plugins.defaults import FAKE_MERGED_S2_TYPE

export, __all__ = strax.exporter()


@export
class PeaksHighEnergy(Peaks):
    __doc__ = HE_PREAMBLE + Peaks.__doc__
    depends_on = ('peaklets_he', 'peaklet_classification_he', 'merged_s2s_he')
    data_kind = 'peaks_he'
    provides = 'peaks_he'
    __version__ = '0.0.1'
    child_ends_with = '_he'

    def infer_dtype(self):
        return self.deps['peaklets_he'].dtype_for('peaklets')

    def compute(self, peaklets_he, merged_s2s_he):
        return super().compute(peaklets_he, merged_s2s_he)



import numpy as np
import straxen
import numba
import typing as ty

import strax

export, __all__ = strax.exporter()


@export
class nVETOEventPositions(strax.Plugin):
    """
    Plugin which computes the interaction position in the nveto as an
    azimuthal angle.
    """
    __version__ = '0.1.1'

    depends_on = ('events_nv', 'hitlets_nv')
    data_kind = 'events_nv'
    provides = 'event_positions_nv'
    compressor = 'zstd'

    position_max_time_nv = straxen.URLConfig(default=20, infer_type=False,
                                             help="Time [ns] within an event use to compute the azimuthal angle of the "
                                                  "event.")

    nveto_pmt_position_map = straxen.URLConfig(
        help="nVeto PMT position mapfile",
        default='resource://nveto_pmt_position.csv?fmt=csv', infer_type=False, )

    def infer_dtype(self):
        return veto_event_positions_dtype()

    def setup(self):
        npmt_pos = self.nveto_pmt_position_map
        # Use records instead of a dataframe.
        self.pmt_properties = npmt_pos.to_records(index=False)

    def compute(self, events_nv, hitlets_nv):
        event_angles = np.zeros(len(events_nv), dtype=self.dtype)

        # Split hitlets by containment, works since we updated event start/end in
        # compute_event_properties.
        hits_in_events = strax.split_by_containment(hitlets_nv, events_nv)

        # Compute hitlets within the first x ns of event:
        hits_in_events, n_prompt = first_hitlets(hits_in_events,
                                                 self.position_max_time_nv)
        event_angles['n_prompt_hitlets'] = n_prompt

        # Compute azimuthal angle and xyz positions:
        angle = get_average_angle(hits_in_events,
                                  self.pmt_properties)
        event_angles['angle'] = angle
        compute_positions(event_angles, hits_in_events, self.pmt_properties)
        strax.copy_to_buffer(events_nv, event_angles, f'_copy_events_nv')

        return event_angles


def veto_event_positions_dtype() -> list:
    dtype = []
    dtype += strax.time_fields
    dtype += [(('Number of prompt hitlets within the first "position_max_time_nv" ns of the event.',
                'n_prompt_hitlets'), np.int16),
              (('Azimuthal angle, where the neutron capture was detected in [0, 2 pi).',
                'angle'), np.float32),
              (('Area weighted mean of position in x [mm]', 'pos_x'), np.float32),
              (('Area weighted mean of position in y [mm]', 'pos_y'), np.float32),
              (('Area weighted mean of position in z [mm]', 'pos_z'), np.float32),
              (('Weighted variance of position in x [mm]', 'pos_x_spread'), np.float32),
              (('Weighted variance of position in y [mm]', 'pos_y_spread'), np.float32),
              (('Weighted variance of position in z [mm]', 'pos_z_spread'), np.float32)
              ]
    return dtype


@numba.njit(cache=True, nogil=True)
def compute_positions(event_angles: np.ndarray,
                      contained_hitlets: numba.typed.typedlist.List,
                      pmt_pos: np.ndarray,
                      start_channel: int = 2000):
    """
    Function which computes some artificial event position for a given
    neutron/muon-veto event. The position is computed based on a simple
    area weighted mean. Please note that the event position can be
    reconstructed in unphysical regions like being within the TPC.

    :param event_angles: Result array of the veto_event_position dtype.
        The result is updated inplace.
    :param contained_hitlets: Hitlets contained in each event.
    :param pmt_pos: Position of the veto PMTs
    :param start_channel: Starting channel of the detector.
    """
    for e_angles, hitlets in zip(event_angles, contained_hitlets):
        prompt_event_area = np.sum(hitlets['area'])
        if prompt_event_area:
            ch = hitlets['channel'] - start_channel
            pos_x = pmt_pos['x'][ch]
            pos_y = pmt_pos['y'][ch]
            pos_z = pmt_pos['z'][ch]

            e_angles['pos_x'] = np.sum(pos_x * hitlets['area']) / prompt_event_area
            e_angles['pos_y'] = np.sum(pos_y * hitlets['area']) / prompt_event_area
            e_angles['pos_z'] = np.sum(pos_z * hitlets['area']) / prompt_event_area
            w = hitlets['area'] / prompt_event_area  # normalized weights
            if len(hitlets) and np.sum(w) > 0:
                e_angles['pos_x_spread'] = np.sqrt(
                    np.sum(w * (pos_x - e_angles['pos_x']) ** 2) / np.sum(w)
                )
                e_angles['pos_y_spread'] = np.sqrt(
                    np.sum(w * (pos_y - e_angles['pos_y']) ** 2) / np.sum(w)
                )
                e_angles['pos_z_spread'] = np.sqrt(
                    np.sum(w * (pos_z - e_angles['pos_z']) ** 2) / np.sum(w)
                )


@numba.njit(cache=True, nogil=True)
def get_average_angle(hitlets_in_event: numba.typed.typedlist.List,
                      pmt_properties: np.ndarray,
                      start_channel: int = 2000, ) -> np.ndarray:
    """
    Computes azimuthal angle as an area weighted mean over all hitlets.

    :param hitlets_in_event: numba.typed.List containing the hitlets per
        event.
    :param pmt_properties: numpy.sturctured.array containing the PMT
        positions in the fields "x" and "y".
    :param start_channel: First channel e.g. 2000 for nevto.
    :return: np.array holding the azimuthal angles.
    """
    res = np.zeros(len(hitlets_in_event), np.float32)
    for ind, hitlets in enumerate(hitlets_in_event):
        if np.sum(hitlets['area']):
            x = pmt_properties['x'][hitlets['channel'] - start_channel]
            y = pmt_properties['y'][hitlets['channel'] - start_channel]

            weighted_mean_x = np.sum(x * hitlets['area']) / np.sum(hitlets['area'])
            weighted_mean_y = np.sum(y * hitlets['area']) / np.sum(hitlets['area'])
            res[ind] = _circ_angle(weighted_mean_x, weighted_mean_y)
        else:
            res[ind] = np.nan
    return res


@numba.njit(cache=True, nogil=True)
def circ_angle(x_values: np.ndarray,
               y_values: np.ndarray) -> np.ndarray:
    """
    Loops over a set of x and y values and computes azimuthal angle.

    :param x_values: x-coordinates
    :param y_values: y-coordinates
    :return: angles
    """
    res = np.zeros(len(x_values), dtype=np.float32)
    for ind, (x, y) in enumerate(zip(x_values, y_values)):
        res[ind] = _circ_angle(x, y)
    return res


@numba.njit(cache=True, nogil=True)
def _circ_angle(x: float, y: float) -> float:
    if x > 0 and y >= 0:
        # 1st quadrant
        angle = np.abs(np.arctan(y / x))
        return angle
    elif x <= 0 and y > 0:
        # 2nd quadrant
        angle = np.abs(np.arctan(x / y))
        return np.pi / 2 + angle
    elif x < 0 and y <= 0:
        # 3rd quadrant
        angle = np.abs(np.arctan(y / x))
        return np.pi + angle
    elif y < 0 and x >= 0:
        # 4th quadrant
        angle = np.abs(np.arctan(x / y))
        return 3 / 2 * np.pi + angle
    elif x == 0 and y == 0:
        return np.NaN
    else:
        print(x, y)
        raise ValueError('It should be impossible to arrive here, '
                         'but somehow we managed.')


@numba.njit(cache=True, nogil=True)
def first_hitlets(hitlets_per_event: np.ndarray,
                  max_time: int) -> ty.Tuple[numba.typed.List, np.ndarray]:
    """
    Returns hitlets within the first "max_time" ns of an event.

    :param hitlets_per_event: numba.typed.List of hitlets per event.
    :param max_time: int max allowed time difference to leading hitlet
        in ns.
    """
    res_hitlets_in_event = numba.typed.List()
    res_n_prompt = np.zeros(len(hitlets_per_event), np.int16)
    for ind, hitlets in enumerate(hitlets_per_event):
        m = (hitlets['time'] - hitlets[0]['time']) < max_time
        h = hitlets[m]
        res_hitlets_in_event.append(h)
        res_n_prompt[ind] = len(h)
    return res_hitlets_in_event, res_n_prompt



"""Some shared defaults"""
DEFAULT_POSREC_ALGO = 'mlp'

HE_PREAMBLE = """High energy channels: attenuated signals of the top PMT-array\n"""

MV_PREAMBLE = 'Muno-Veto Plugin: Same as the corresponding nVETO-PLugin.\n'

NV_HIT_DEFAULTS = {
    'save_outside_hits_nv': (3, 15),
    'hit_min_amplitude_nv': 'cmt://hit_thresholds_nv?version=ONLINE&run_id=plugin.run_id',
}

MV_HIT_DEFAULTS = {
    'save_outside_hits_mv': (2, 5),
    'hit_min_amplitude_mv': 'cmt://hit_thresholds_mv?version=ONLINE&run_id=plugin.run_id',
}

FAKE_MERGED_S2_TYPE = -42



import strax
import numpy as np
import straxen

export, __all__ = strax.exporter()


@export
class IndividualPeakMonitor(strax.Plugin):
    """
    Plugin to write data needed for the online SE monitor to the 
    online-monitor collection in the runs-database. Data that is written by
    this plugin should be small such as to not overload the runs-
    database. If the peaks are large, random
    max_bytes of data are selected from the peaks.

    This plugin takes 'peak_basics' and 'peak_positions_mlp'. Although 
    they are not strictly related, they are aggregated into a single data_type
    in order to minimize the number of documents in the online monitor.

    Produces 'individual_peak_monitor' with info on the peaks and their
    positions.
    """

    online_max_bytes = straxen.URLConfig(
        default=6e6,
        track=True,
        help='Maximum amount of bytes of data for MongoDB document'
    )

    depends_on = ('peak_basics', 'peak_positions_mlp')
    provides = 'individual_peak_monitor'
    data_kind = 'individual_peak_monitor'
    __version__ = '0.0.1'

    def infer_dtype(self):
        dtype = [
                    (('Peak integral in PE', 'area'),
                     np.float32),
                    (('Reconstructed mlp peak x-position', 'x_mlp'),
                     np.float32),
                    (('Reconstructed mlp peak y-position', 'y_mlp'),
                     np.float32),
                    (('Width (in ns) of the central 50% area of the peak', 'range_50p_area'),
                     np.float32),
                    (('Fraction of original peaks array length that is saved', 'weight'),
                     np.float32),
                ] + strax.time_fields
        return dtype

    def compute(self, peaks):
        peaks_size = peaks.nbytes

        if peaks_size > self.online_max_bytes:
            # Calculate fraction of the data that can be kept
            # to reduce datasize
            new_len = int(len(peaks) / peaks_size * self.online_max_bytes)
            idx = np.random.choice(np.arange(len(peaks)), replace=False, size=new_len)
            data = peaks[np.sort(idx)]

        else:  # peaks_size <= self.max_bytes:
            data = peaks
        res = np.zeros(len(data), dtype=self.dtype)
        res['time'] = data['time']
        res['x_mlp'] = data['x_mlp']
        res['y_mlp'] = data['y_mlp']
        res['area'] = data['area']
        res['range_50p_area'] = data['range_50p_area']
        res['endtime'] = data['endtime']

        if len(data):
            res['weight'] = len(peaks) / len(data)
        else:
            res['weight'] = 0

        return res



import numpy as np
import straxen
from straxen.plugins.defaults import DEFAULT_POSREC_ALGO

import strax

export, __all__ = strax.exporter()


@export
class EventPositions(strax.Plugin):
    """
    Computes the observed and corrected position for the main S1/S2
    pairs in an event. For XENONnT data, it returns the FDC corrected
    positions of the default_reconstruction_algorithm. In case the fdc_map
    is given as a file (not through CMT), then the coordinate system
    should be given as (x, y, z), not (x, y, drift_time).
    """

    depends_on = ('event_basics',)

    __version__ = '0.3.0'

    default_reconstruction_algorithm = straxen.URLConfig(
        default=DEFAULT_POSREC_ALGO,
        help="default reconstruction algorithm that provides (x,y)"
    )

    electron_drift_velocity = straxen.URLConfig(
        default='cmt://'
                'electron_drift_velocity'
                '?version=ONLINE&run_id=plugin.run_id',
        cache=True,
        help='Vertical electron drift velocity in cm/ns (1e4 m/ms)'
    )

    electron_drift_time_gate = straxen.URLConfig(
        default='cmt://'
                'electron_drift_time_gate'
                '?version=ONLINE&run_id=plugin.run_id',
        help='Electron drift time from the gate in ns',
        cache=True)

    fdc_map = straxen.URLConfig(
        infer_type=False,
        help='3D field distortion correction map path',
        default='legacy-fdc://xenon1t_sr0_sr1?run_id=plugin.run_id')

    z_bias_map = straxen.URLConfig(
        infer_type=False,
        help='Map of Z bias due to non uniform drift velocity/field',
        default='legacy-z_bias://0')

    def infer_dtype(self):
        dtype = []
        for j in 'x y r'.split():
            comment = f'Main interaction {j}-position, field-distortion corrected (cm)'
            dtype += [(j, np.float32, comment)]
            for s_i in [1, 2]:
                comment = f'Alternative S{s_i} interaction (rel. main S{3 - s_i}) {j}-position, field-distortion corrected (cm)'
                field = f'alt_s{s_i}_{j}_fdc'
                dtype += [(field, np.float32, comment)]

        for j in ['z']:
            comment = 'Interaction z-position, corrected to non-uniform drift velocity (cm)'
            dtype += [(j, np.float32, comment)]
            comment = 'Interaction z-position, corrected to non-uniform drift velocity, duplicated (cm)'
            dtype += [(j + "_dv_corr", np.float32, comment)]
            for s_i in [1, 2]:
                comment = f'Alternative S{s_i} z-position (rel. main S{3 - s_i}), corrected to non-uniform drift velocity (cm)'
                field = f'alt_s{s_i}_z'
                dtype += [(field, np.float32, comment)]
                # values for corrected Z position
                comment = f'Alternative S{s_i} z-position (rel. main S{3 - s_i}), corrected to non-uniform drift velocity, duplicated (cm)'
                field = f'alt_s{s_i}_z_dv_corr'
                dtype += [(field, np.float32, comment)]


        naive_pos = []
        fdc_pos = []
        for j in 'r z'.split():
            naive_pos += [(f'{j}_naive',
                           np.float32,
                           f'Main interaction {j}-position with observed position (cm)')]
            fdc_pos += [(f'{j}_field_distortion_correction',
                         np.float32,
                         f'Correction added to {j}_naive for field distortion (cm)')]
            for s_i in [1, 2]:
                naive_pos += [(
                    f'alt_s{s_i}_{j}_naive',
                    np.float32,
                    f'Alternative S{s_i} interaction (rel. main S{3 - s_i}) {j}-position with observed position (cm)')]
                fdc_pos += [(f'alt_s{s_i}_{j}_field_distortion_correction',
                             np.float32,
                             f'Correction added to alt_s{s_i}_{j}_naive for field distortion (cm)')]
        dtype += naive_pos + fdc_pos
        for s_i in [1, 2]:
            dtype += [(f'alt_s{s_i}_theta',
                       np.float32,
                       f'Alternative S{s_i} (rel. main S{3 - s_i}) interaction angular position (radians)')]

        dtype += [('theta', np.float32, f'Main interaction angular position (radians)')]
        return dtype + strax.time_fields

    def setup(self):
        self.coordinate_scales = [1., 1., - self.electron_drift_velocity]
        self.map = self.fdc_map

    def compute(self, events):

        result = {'time': events['time'],
                  'endtime': strax.endtime(events)}

        # s_i == 0 indicates the main event, while s_i != 0 means alternative S1 or S2 is used based on s_i value
        # while the other peak is the main one (e.g., s_i == 1 means that the event is defined using altS1 and main S2)
        for s_i in [0, 1, 2]:
            # alt_sx_interaction_drift_time is calculated between main Sy and alternative Sx
            drift_time = events['drift_time'] if not s_i else events[f'alt_s{s_i}_interaction_drift_time']
            z_obs = - self.electron_drift_velocity * drift_time
            xy_pos = 's2_' if s_i != 2 else 'alt_s2_'
            orig_pos = np.vstack([events[f'{xy_pos}x'], events[f'{xy_pos}y'], z_obs]).T
            r_obs = np.linalg.norm(orig_pos[:, :2], axis=1)
            z_obs = z_obs + self.electron_drift_velocity * self.electron_drift_time_gate

            # apply z bias correction
            z_dv_delta = self.z_bias_map(np.array([r_obs, z_obs]).T, map_name='z_bias_map')
            corr_pos = np.vstack([events[f"{xy_pos}x"], events[f"{xy_pos}y"], z_obs - z_dv_delta]).T
            # apply r bias correction
            delta_r = self.map(corr_pos)
            with np.errstate(invalid='ignore', divide='ignore'):
                r_cor = r_obs + delta_r
                scale = np.divide(r_cor, r_obs, out=np.zeros_like(r_cor), where=r_obs != 0)

            pre_field = '' if s_i == 0 else f'alt_s{s_i}_'
            post_field = '' if s_i == 0 else '_fdc'
            result.update({f'{pre_field}x{post_field}': orig_pos[:, 0] * scale,
                           f'{pre_field}y{post_field}': orig_pos[:, 1] * scale,
                           f'{pre_field}r{post_field}': r_cor,
                           f'{pre_field}r_naive': r_obs,
                           f'{pre_field}r_field_distortion_correction': delta_r,
                           f'{pre_field}theta': np.arctan2(orig_pos[:, 1], orig_pos[:, 0]),
                           f'{pre_field}z_naive': z_obs,
                           # using z_dv_corr (z_obs - z_dv_delta) in agreement with the dtype description
                           # the FDC for z (z_cor) is found to be not reliable (see #527)
                           f'{pre_field}z': z_obs - z_dv_delta,
                           f'{pre_field}z_dv_corr': z_obs - z_dv_delta,
                           })
        return result



import numpy as np
import strax

export, __all__ = strax.exporter()


@export
class EventAmbience(strax.Plugin):
    """
    Save Ambience of the main S1 and main S2 in the event.
    References:
        * v0.0.4 reference: xenon:xenonnt:ac:prediction:shadow_ambience
    """
    __version__ = '0.0.4'
    depends_on = ('event_basics', 'peak_basics', 'peak_ambience')
    provides = 'event_ambience'

    @property
    def origin_dtype(self):
        return ['lh_before', 's0_before', 's1_before', 's2_before', 's2_near']

    def infer_dtype(self):
        dtype = []
        for ambience in self.origin_dtype:
            dtype.append(((f"Number of  {' '.join(ambience.split('_'))} main S1",
                           f's1_n_{ambience}'), np.int16))
            dtype.append(((f"Number of  {' '.join(ambience.split('_'))} main S2",
                           f's2_n_{ambience}'), np.int16))
        dtype += strax.time_fields
        return dtype

    def compute(self, events, peaks):
        split_peaks = strax.split_by_containment(peaks, events)

        # 1. Initialization, ambience is set to be the lowest possible value
        result = np.zeros(len(events), self.dtype)
        # 2. Assign peaks features to main S1, main S2 in the event
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            for idx, main_peak in zip([event['s1_index'], event['s2_index']], ['s1_', 's2_']):
                if idx >= 0:
                    for ambience in self.origin_dtype:
                        result[f'{main_peak}n_{ambience}'][event_i] = sp[f'n_{ambience}'][idx]

        # 3. Set time and endtime for events
        result['time'] = events['time']
        result['endtime'] = strax.endtime(events)
        return result



from immutabledict import immutabledict
import strax
import straxen
from straxen.common import pax_file, aux_repo


def get_x1t_context_config():
    """wrapper to get straxen.contexts after imports"""
    from straxen.contexts import common_opts

    x1t_context_config = {
        **common_opts,
        **dict(
            check_available=('raw_records', 'records', 'peaklets',
                             'events', 'event_info'),
            free_options=('channel_map',),
            use_per_run_defaults=True,
            store_run_fields=tuple(
                [x for x in common_opts['store_run_fields'] if x != 'mode']
                + ['trigger.events_built', 'reader.ini.name']))}
    x1t_context_config.update(
        dict(register=common_opts['register'] +
                      [straxen.PeakPositions1T,
                       straxen.RecordsFromPax,
                       straxen.EventInfo1T,
                       ]))
    return x1t_context_config


x1t_common_config = dict(
    check_raw_record_overlaps=False,
    allow_sloppy_chunking=True,
    n_tpc_pmts=248,
    n_top_pmts=127,
    channel_map=immutabledict(
        # (Minimum channel, maximum channel)
        tpc=(0, 247),
        diagnostic=(248, 253),
        aqmon=(254, 999)),
    # Records
    hev_gain_model='cmt://to_pe_model?version=v1&detector=1t&run_id=plugin.run_id',
    pmt_pulse_filter=(
        0.012, -0.119,
        2.435, -1.271, 0.357, -0.174, -0., -0.036,
        -0.028, -0.019, -0.025, -0.013, -0.03, -0.039,
        -0.005, -0.019, -0.012, -0.015, -0.029, 0.024,
        -0.007, 0.007, -0.001, 0.005, -0.002, 0.004, -0.002),
    hit_min_amplitude='legacy-thresholds://XENON1T_SR1',
    tail_veto_threshold=int(1e5),
    save_outside_hits=(3, 3),
    # Peaklets
    peaklet_gap_threshold=350,
    gain_model='cmt://to_pe_model?version=v1&detector=1t&run_id=plugin.run_id',
    peak_split_gof_threshold=(
        None,  # Reserved
        ((0.5, 1), (3.5, 0.25)),
        ((2, 1), (4.5, 0.4))),
    peak_min_pmts=2,
    # MergedS2s
    s2_merge_gap_thresholds=((1.7, 5.0e3), (4.0, 500.), (5.0, 0.)),
    # Peaks
    # Smaller right extension since we applied the filter
    peak_right_extension=30,
    s1_max_rise_time_post100=150,
    s1_min_coincidence=3,
    event_s1_min_coincidence=3,
    # Events*
    left_event_extension=int(0.3e6),
    right_event_extension=int(1e6),
    elife=1e6,
    electron_drift_velocity=1.3325e-4,
    max_drift_length=96.9,
    electron_drift_time_gate=1700,
    se_gain=28.2,
    avg_se_gain=28.2,
    rel_extraction_eff=1.0,
    rel_light_yield=1.0,
    s1_xyz_map=f'itp_map://resource://{pax_file("XENON1T_s1_xyz_lce_true_kr83m_SR1_pax-680_fdc-3d_v0.json")}?fmt=json',  # noqa
    s2_xy_map=f'itp_map://resource://{pax_file("XENON1T_s2_xy_ly_SR1_v2.2.json")}?fmt=json',
    g1=0.1426,
    g2=11.55 / (1 - 0.63),
)


def demo():
    """Return strax context used in the straxen demo notebook"""
    straxen.download_test_data()

    st = strax.Context(
        storage=[strax.DataDirectory('./strax_data'),
                 strax.DataDirectory('./strax_test_data',
                                     deep_scan=True,
                                     provide_run_metadata=True,
                                     readonly=True)],
        forbid_creation_of=straxen.daqreader.DAQReader.provides,
        config=dict(**x1t_common_config),
        **get_x1t_context_config())

    # Use configs that are always available
    st.set_config(dict(
        hev_gain_model='legacy-to-pe://1T_to_pe_placeholder',
        gain_model='legacy-to-pe://1T_to_pe_placeholder',
        elife=1e6,
        electron_drift_velocity=1.3325e-4,
        se_gain=28.2,
        avg_se_gain=28.2,
        rel_extraction_eff=1.0,
        s1_xyz_map=f'itp_map://resource://{pax_file("XENON1T_s1_xyz_lce_true_kr83m_SR1_pax-680_fdc-3d_v0.json")}?fmt=json',
        s2_xy_map=f'itp_map://resource://{pax_file("XENON1T_s2_xy_ly_SR1_v2.2.json")}?fmt=json',
    ))
    return st


def fake_daq():
    """Context for processing fake DAQ data in the current directory"""
    st = strax.Context(
        storage=[strax.DataDirectory('./strax_data'),
                 # Fake DAQ puts run doc JSON in same folder:
                 strax.DataDirectory('./from_fake_daq',
                                     provide_run_metadata=True,
                                     readonly=True)],
        config=dict(daq_input_dir='./from_fake_daq',
                    daq_chunk_duration=int(2e9),
                    daq_compressor='lz4',
                    n_readout_threads=8,
                    daq_overlap_chunk_duration=int(2e8),
                    **x1t_common_config),
        **get_x1t_context_config())
    st.register(straxen.Fake1TDAQReader)
    return st


def xenon1t_dali(output_folder='./strax_data', build_lowlevel=False, **kwargs):
    context_options = {
        **get_x1t_context_config(),
        **kwargs}

    st = strax.Context(
        storage=[
            strax.DataDirectory(
                '/dali/lgrandi/xenon1t/strax_converted/raw',
                take_only='raw_records',
                provide_run_metadata=True,
                readonly=True),
            strax.DataDirectory(
                '/dali/lgrandi/xenon1t/strax_converted/processed',
                readonly=True),
            strax.DataDirectory(output_folder)],
        config=dict(**x1t_common_config),
        # When asking for runs that don't exist, throw an error rather than
        # starting the pax converter
        forbid_creation_of=(
            straxen.daqreader.DAQReader.provides if build_lowlevel
            else straxen.daqreader.DAQReader.provides + ('records', 'peaklets')),
        **context_options)
    return st


def xenon1t_led(**kwargs):
    st = xenon1t_dali(**kwargs)
    st.set_context_config(
        {'check_available': ('raw_records', 'led_calibration'),
         'free_options': list(get_x1t_context_config().keys())
         })
    # Return a new context with only raw_records and led_calibration registered
    st = st.new_context(
        replace=True,
        config=st.config,
        storage=st.storage,
        **st.context_config)
    st.register([straxen.RecordsFromPax, straxen.LEDCalibration])
    return st


def xenon1t_simulation(output_folder='./strax_data'):
    import wfsim
    st = strax.Context(
        storage=strax.DataDirectory(output_folder),
        config=dict(
            fax_config='fax_config_1t.json',
            detector='XENON1T',
            **x1t_common_config),
        **get_x1t_context_config())
    st.register(wfsim.RawRecordsFromFax1T)
    st.deregister_plugins_with_missing_dependencies()
    return st



import strax
import numpy as np
from straxen.plugins.online_monitor_nv.online_monitor_nv import OnlineMonitorNV, veto_monitor_dtype

export, __all__ = strax.exporter()


@export
class OnlineMonitorMV(OnlineMonitorNV):
    __doc__ = OnlineMonitorNV.__doc__.replace('_nv', '_mv').replace('nVeto', 'muVeto')
    depends_on = ('hitlets_mv', 'events_mv')
    provides = 'online_monitor_mv'
    data_kind = 'online_monitor_mv'

    # Needed in case we make again an muVETO child.
    ends_with = '_mv'
    child_plugin = True

    __version__ = '0.0.2'

    def infer_dtype(self):
        self.channel_range = self.channel_map['mv']
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1
        return veto_monitor_dtype(self.ends_with, self.n_channel, self.events_area_nbins)

    def compute(self, hitlets_mv, events_mv, start, end):
        events_mv = np.copy(events_mv)
        return super().compute(hitlets_mv, events_mv, start, end)



import numpy as np
import strax

export, __all__ = strax.exporter()


@export
class DistinctChannels(strax.LoopPlugin):
    """
    Compute the number of contributing PMTs that contribute to the
    alt_s1 but not to the main S1.
    """
    __version__ = '0.1.1'
    depends_on = ('event_basics', 'peaks')
    loop_over = 'events'
    dtype = [
                ('alt_s1_distinct_channels',
                 np.int32,
                 'Number of PMTs contributing to the secondary S1 '
                 'that do not contribute to the main S1'), ] + strax.time_fields

    def compute_loop(self, event, peaks):
        if event['alt_s1_index'] == -1:
            n_distinct = 0
        else:
            s1_a = peaks[event['s1_index']]
            s1_b = peaks[event['alt_s1_index']]
            s1_a_peaks = np.nonzero((s1_a['area_per_channel'] > 0) * 1)
            s1_b_peaks = np.nonzero((s1_b['area_per_channel'] > 0) * 1)
            n_distinct = 0
            for channel in range(len(s1_b_peaks[0])):
                if s1_b_peaks[0][channel] not in s1_a_peaks[0]:
                    n_distinct += 1

        return dict(alt_s1_distinct_channels=n_distinct,
                    time=event['time'],
                    endtime=event['endtime'])



import strax
import straxen
import tarfile
import io
import os
from warnings import warn
from os import environ as os_environ
from immutabledict import immutabledict
from importlib import import_module
import numpy as np


export, __all__ = strax.exporter()


nt_test_run_id = '012882'


@export
def download_test_data(test_data='https://raw.githubusercontent.com/XENONnT/strax_auxiliary_files/353b2c60a01e96f67e4ba544ce284bd91241964d/strax_files/strax_test_data_straxv1.1.0.tar',  #  noqa
                       ):
    """Downloads strax test data to strax_test_data in the current directory"""
    blob = straxen.common.get_resource(test_data, fmt='binary')
    f = io.BytesIO(blob)
    tf = tarfile.open(fileobj=f)
    tf.extractall()


@export
def _overwrite_testing_function_file(function_file):
    """For testing purposes allow this function file to be loaded from HOME/testing_folder"""
    if not _is_on_pytest():
        # If we are not on a pytest, never try using a local file.
        return function_file

    home = os.environ.get('HOME')
    if home is None:
        # Impossible to load from non-existent folder
        return function_file

    testing_file = os.path.join(home, function_file)

    if os.path.exists(testing_file):
        # For testing purposes allow loading from 'home/testing_folder'
        warn(f'Using local function: {function_file} from {testing_file}! '
             f'If you are not integrated testing on github you should '
             f'absolutely remove this file. (See #559)')
        function_file = testing_file

    return function_file


def is_installed(module):
    try:
        import_module(module)
        return True
    except ModuleNotFoundError:
        return False


@export
def _is_on_pytest():
    """Check if we are on a pytest"""
    return 'PYTEST_CURRENT_TEST' in os_environ


def _get_fake_daq_reader():
    class DAQReader(straxen.DAQReader):
        """
        Dummy version of the DAQ reader to make sure that all the testing
        data produced here will have a different lineage
        """
        __version__ = "MOCKTESTDATA"
    return DAQReader


def nt_test_context(target_context='xenonnt_online',
                    deregister=(),
                    keep_default_storage=False,
                    **kwargs) -> strax.Context:
    """
    Get a dummy context with full nt-like data simulated data (except aqmon)
    to allow testing plugins

    :param target_context: Which contexts from straxen.contexts to test
    :param deregister: a list of plugins from the context
    :param keep_default_storage: if to True, keep the default context
        storage. Usually, you don't need this since all the data will be
         stored in a separate test data folder.
    :param kwargs: Any kwargs are passed to the target-context
    :return: a context
    """
    if not straxen.utilix_is_configured(warning_message=False):
        kwargs.setdefault('_database_init', False)

    st = getattr(straxen.contexts, target_context)(**kwargs)
    st.set_config({
        'diagnose_sorting': True, 'diagnose_overlapping': True, 'store_per_channel': True})
    st.register(_get_fake_daq_reader())
    download_test_data('https://raw.githubusercontent.com/XENONnT/'
                       'strax_auxiliary_files/'
                       'f0d177401e11408b273564f0e29df77528e83d26/'
                       'strax_files/'
                       '012882-raw_records-z7q2d2ye2t.tar')
    if keep_default_storage:
        st.storage += [strax.DataDirectory('./strax_test_data')]
    else:
        st.storage = [strax.DataDirectory('./strax_test_data')]
    assert st.is_stored(nt_test_run_id, 'raw_records'), os.listdir(st.storage[-1].path)

    to_remove = list(deregister)
    for plugin in to_remove:
        del st._plugin_class_registry[plugin]
    return st


def create_unique_intervals(size, time_range=(0, 40), allow_zero_length=True):
    """
    Hypothesis stragtegy which creates unqiue time intervals.

    :param size: Number of intervals desired. Can be less if non-unique
        intervals are found.
    :param time_range: Time range in which intervals should be.
    :param allow_zero_length: If true allow zero length intervals.
    """
    from hypothesis import strategies
    strat = strategies.lists(elements=strategies.integers(*time_range),
                             min_size=size * 2,
                             max_size=size * 2
                             ).map(lambda x: _convert_to_interval(x, allow_zero_length))
    return strat


def _convert_to_interval(time_stamps, allow_zero_length):
    time_stamps = np.sort(time_stamps)
    intervals = np.zeros(len(time_stamps) // 2, strax.time_dt_fields)
    intervals['dt'] = 1
    intervals['time'] = time_stamps[::2]
    intervals['length'] = time_stamps[1::2] - time_stamps[::2]

    if not allow_zero_length:
        intervals = intervals[intervals['length'] > 0]
    return np.unique(intervals)


@strax.takes_config(
    strax.Option('secret_time_offset', default=0, track=False),
    strax.Option('recs_per_chunk', default=10, track=False),
    strax.Option('n_chunks', default=2, track=False,
                 help='Number of chunks for the dummy raw records we are writing here'),
    strax.Option('channel_map', track=False, type=immutabledict,
                 help="frozendict mapping subdetector to (min, max) "
                      "channel number.")
)
class DummyRawRecords(strax.Plugin):
    """
    Provide dummy raw records for the mayor raw_record types
    """
    provides = straxen.daqreader.DAQReader.provides
    parallel = 'process'
    depends_on = tuple()
    data_kind = immutabledict(zip(provides, provides))
    rechunk_on_save = False
    dtype = {p: strax.raw_record_dtype() for p in provides}

    def setup(self):
        self.channel_map_keys = {'he': 'he',
                                 'nv': 'nveto',
                                 'aqmon': 'aqmon',
                                 'aux_mv': 'aux_mv',
                                 's_mv': 'mv',
                                 }  # s_mv otherwise same as aux in endswith

    def source_finished(self):
        return True

    def is_ready(self, chunk_i):
        return chunk_i < self.config['n_chunks']

    def compute(self, chunk_i):
        t0 = chunk_i + self.config['secret_time_offset']
        if chunk_i < self.config['n_chunks'] - 1:
            # One filled chunk
            r = np.zeros(self.config['recs_per_chunk'], self.dtype['raw_records'])
            r['time'] = t0
            r['length'] = r['dt'] = 1
            r['channel'] = np.arange(len(r))
        else:
            # One empty chunk
            r = np.zeros(0, self.dtype['raw_records'])

        res = {}
        for p in self.provides:
            rr = np.copy(r)
            # Add detector specific channel offset:
            for key, channel_key in self.channel_map_keys.items():
                if channel_key not in self.config['channel_map']:
                    # Channel map for 1T is different.
                    continue
                if p.endswith(key):
                    first_channel, last_channel = self.config['channel_map'][channel_key]
                    rr['channel'] += first_channel
                    if key == 'aqmon':
                        # explicitly clip these channels as we have an additional check higher in the chain
                        first_channel=int(min(straxen.AqmonChannels))
                        last_channel=int(max(straxen.AqmonChannels))

                    rr = rr[(rr['channel']>=first_channel) & (rr['channel']<last_channel)]
            res[p] = self.chunk(start=t0, end=t0 + 1, data=rr, data_type=p)
        return res



import strax
import numpy as np


export, __all__ = strax.exporter()


@export
class EventPeaks(strax.Plugin):
    """
    Add event number for peaks and drift times of all s2 depending on the largest s1.
    Link - https://xe1t-wiki.lngs.infn.it/doku.php?id=weiss:analysis:ms_plugin
    """
    __version__ = '0.0.1'
    depends_on = ('event_basics', 'peak_basics', 'peak_positions')
    provides = 'peak_per_event'
    data_kind = 'peaks'
    save_when = strax.SaveWhen.TARGET

    def infer_dtype(self):
        dtype = strax.time_fields + [
            ('drift_time', np.float32, 'Drift time between main S1 and S2 in ns'),
            ('event_number', np.int64, 'Event number in this dataset'),
        ]
        return dtype

    def compute(self, events, peaks):
        split_peaks = strax.split_by_containment(peaks, events)
        split_peaks_ind = strax.fully_contained_in(peaks, events)
        result = np.zeros(len(peaks), self.dtype)
        result.fill(np.nan)

        # Assign peaks features to main S1 and main S2 in the event
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            result['drift_time'][split_peaks_ind==event_i] = sp['center_time'] - event['s1_center_time']
        result['event_number'] = split_peaks_ind
        result['drift_time'][peaks['type'] != 2] = np.nan
        result['time'] = peaks['time']
        result['endtime'] = strax.endtime(peaks)
        return result



import strax

export, __all__ = strax.exporter()


@export
class EventInfo1T(strax.MergeOnlyPlugin):
    """
    Plugin which merges the information of all event data_kinds into a
    single data_type.

    This only uses 1T data-types as several event-plugins are nT only
    """
    depends_on = ['event_basics',
                  'event_positions',
                  'corrected_areas',
                  'energy_estimates',
                  ]
    provides = 'event_info'
    save_when = strax.SaveWhen.ALWAYS
    __version__ = '0.0.1'



import numpy as np
import os
import glob

import strax

export, __all__ = strax.exporter()


def records_needed(pulse_length, samples_per_record):
    """Return records needed to store pulse_length samples"""
    return np.ceil(pulse_length / samples_per_record).astype(np.int64)


@export
def pax_to_records(input_filename,
                   samples_per_record=strax.DEFAULT_RECORD_LENGTH,
                   events_per_chunk=10):
    """Return pulse records array from pax zip input_filename
    Convert pax .zip files to flat records format
    This only works if you have pax installed in your strax environment,
    which is somewhat tricky.
    """

    # Monkeypatch matplotlib so pax is importable
    # See https://github.com/XENON1T/pax/pull/734
    import matplotlib
    matplotlib._cntr = None

    from pax import core  # Pax is not a dependency

    mypax = core.Processor('XENON1T', config_dict=dict(
        pax=dict(
            look_for_config_in_runs_db=False,
            plugin_group_names=['input'],
            encoder_plugin=None,
            input_name=input_filename),
        # Fast startup: skip loading big maps
        WaveformSimulator=dict(
            s1_light_yield_map='placeholder_map.json',
            s2_light_yield_map='placeholder_map.json',
            s1_patterns_file=None,
            s2_patterns_file=None)))

    print(f"Starting conversion, {events_per_chunk} evt/chunk")

    results = []

    def finish_results():
        nonlocal results
        records = np.concatenate(results)
        # In strax data, records are always stored
        # sorted, baselined and integrated
        records = strax.sort_by_time(records)
        print("Returning %d records" % len(records))
        results = []
        return records

    for event in mypax.get_events():
        event = mypax.process_event(event)

        if not len(event.pulses):
            # Triggerless pax data contains many empty events
            # at the end. With the fixed events per chunk setting
            # this can lead to empty files, which confuses strax.
            continue

        pulse_lengths = np.array([p.length
                                  for p in event.pulses])

        n_records_tot = records_needed(pulse_lengths,
                                       samples_per_record).sum()
        records = np.zeros(n_records_tot,
                           dtype=strax.raw_record_dtype(samples_per_record))
        output_record_index = 0  # Record offset in data

        for p in event.pulses:
            n_records = records_needed(p.length, samples_per_record)

            for rec_i in range(n_records):
                r = records[output_record_index]
                r['time'] = (event.start_time
                             + p.left * 10
                             + rec_i * samples_per_record * 10)
                r['channel'] = p.channel
                r['pulse_length'] = p.length
                r['record_i'] = rec_i
                r['dt'] = 10

                # How much are we storing in this record?
                if rec_i != n_records - 1:
                    # There's more chunks coming, so we store a full chunk
                    n_store = samples_per_record
                    assert p.length > samples_per_record * (rec_i + 1)
                else:
                    # Just enough to store the rest of the data
                    # Note it's not p.length % samples_per_record!!!
                    # (that would be zero if we have to store a full record)
                    n_store = p.length - samples_per_record * rec_i

                assert 0 <= n_store <= samples_per_record
                r['length'] = n_store

                offset = rec_i * samples_per_record
                r['data'][:n_store] = p.raw_data[offset:offset + n_store]
                output_record_index += 1

        results.append(records)
        if len(results) >= events_per_chunk:
            yield finish_results()

    mypax.shutdown()

    if len(results):
        y = finish_results()
        if len(y):
            yield y


@export
@strax.takes_config(
    strax.Option('pax_raw_dir', default='/data/xenon/raw', track=False, infer_type=False,
                 help="Directory with raw pax datasets"),
    strax.Option('stop_after_zips', default=0, track=False, infer_type=False,
                 help="Convert only this many zip files. 0 = all."),
    strax.Option('events_per_chunk', default=50, track=False, infer_type=False,
                 help="Number of events to yield per chunk"),
    strax.Option('samples_per_record', default=strax.DEFAULT_RECORD_LENGTH,
                 track=False, infer_type=False,
                 help="Number of samples per record")
)
class RecordsFromPax(strax.Plugin):
    provides = 'raw_records'
    data_kind = 'raw_records'
    compressor = 'zstd'
    depends_on = tuple()
    parallel = False
    rechunk_on_save = False

    def infer_dtype(self):
        return strax.raw_record_dtype(self.config['samples_per_record'])

    def iter(self, *args, **kwargs):
        if not os.path.exists(self.config['pax_raw_dir']):
            raise FileNotFoundError(self.config['pax_raw_dir'])
        input_dir = os.path.join(self.config['pax_raw_dir'], self.run_id)
        pax_files = sorted(glob.glob(input_dir + '/XENON*.zip'))
        pax_sizes = np.array([os.path.getsize(x)
                              for x in pax_files])
        print(f"Found {len(pax_files)} files, {pax_sizes.sum() / 1e9:.2f} GB")
        last_endtime = 0

        for file_i, in_fn in enumerate(pax_files):
            if (self.config['stop_after_zips']
                    and file_i >= self.config['stop_after_zips']):
                break
            for records in pax_to_records(
                    in_fn,
                    samples_per_record=self.config['samples_per_record'],
                    events_per_chunk=self.config['events_per_chunk']):

                if not len(records):
                    continue
                if last_endtime == 0:
                    last_endtime = records[0]['time']
                new_endtime = strax.endtime(records).max()

                yield self.chunk(start=last_endtime,
                                 end=new_endtime,
                                 data=records)

                last_endtime = new_endtime



import strax
import straxen
import numpy as np
export, __all__ = strax.exporter()


@export
class EventInfoMS(strax.Plugin):
    """
    Plugin to collect multiple-scatter event observables
    """
    __version__ = '0.0.2'
    depends_on = (
        'event_info',
        'peak_basics', 'peak_per_event', 'peak_corrections', 'peak_positions')
    provides = 'event_ms_naive'
    save_when = strax.SaveWhen.TARGET

    # config options don't double cache things from the resource cache!
    g1 = straxen.URLConfig(
        default='bodega://g1?bodega_version=v2',
        help='S1 gain in PE / photons produced',
    )
    g2 = straxen.URLConfig(
        default='bodega://g2?bodega_version=v2',
        help='S2 gain in PE / electrons produced',
    )
    lxe_w = straxen.URLConfig(
        default=13.7e-3,
        help='LXe work function in quanta/keV'
    )
    electron_drift_velocity = straxen.URLConfig(
        default='cmt://'
                'electron_drift_velocity'
                '?version=ONLINE&run_id=plugin.run_id',
        cache=True,
        help='Vertical electron drift velocity in cm/ns (1e4 m/ms)'
    )
    max_drift_length = straxen.URLConfig(
        default=straxen.tpc_z, infer_type=False,
        help='Total length of the TPC from the bottom of gate to the '
             'top of cathode wires [cm]')

    ms_window_fac = straxen.URLConfig(
        default=1.01, type=(int, float),
        help='Max drift time window to look for peaks in multiple scatter events'
    )

    def infer_dtype(self):
        dtype = strax.time_fields + [
            (('Sum of S1 areas in event',
              's1_sum'), np.float32),
            (('Corrected S1 area based on average position of S2s in event',
              'cs1_multi'), np.float32),
            (('Corrected S1 area based on average position of S2s in event before time-dep LY correction',
              'cs1_multi_wo_timecorr'), np.float32),
            (('Sum of S2 areas in event',
              's2_sum'), np.float32),
            (('Sum of corrected S2 areas in event',
              'cs2_sum'), np.float32),
            (('Sum of corrected S2 areas in event S2 before elife correction',
              'cs2_wo_timecorr_sum'), np.float32),
            (('Sum of corrected S2 areas in event before SEG/EE and elife corrections',
              'cs2_wo_elifecorr_sum'), np.float32),
            (('Average of S2 area fraction top in event',
              'cs2_area_fraction_top_avg'), np.float32),
            (('Sum of the energy estimates in event',
              'ces_sum'), np.float32),
            (('Sum of the charge estimates in event',
              'e_charge_sum'), np.float32),
            (('Average x position of S2s in event',
              'x_avg'), np.float32),
            (('Average y position of S2s in event',
              'y_avg'), np.float32),
            (('Average observed z position of energy deposits in event',
              'z_obs_avg'), np.float32),
            (('Number of S2s in event',
              'multiplicity'), np.int32),
        ]
        return dtype

    def setup(self):
        self.drift_time_max = int(self.max_drift_length / self.electron_drift_velocity)

    def cs1_to_e(self, x):
        return self.lxe_w * x / self.g1

    def cs2_to_e(self, x):
        return self.lxe_w * x / self.g2

    def compute(self, events, peaks):
        split_peaks = strax.split_by_containment(peaks, events)
        result = np.zeros(len(events), self.infer_dtype())

        # Assign peaks features to main S1 and main S2 in the event
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            cond = (sp['type'] == 2) & (sp['drift_time'] > 0)
            cond &= (sp['drift_time'] < self.ms_window_fac * self.drift_time_max) & (sp['cs2'] > 0)
            result['s2_sum'][event_i] = np.nansum(sp[cond]['area'])
            result['cs2_sum'][event_i] = np.nansum(sp[cond]['cs2'])
            result['cs2_wo_timecorr_sum'][event_i] = np.nansum(sp[cond]['cs2_wo_timecorr'])
            result['cs2_wo_elifecorr_sum'][event_i] = np.nansum(sp[cond]['cs2_wo_elifecorr'])         
            result['s1_sum'][event_i] = np.nansum(sp[sp['type'] == 1]['area'])

            if np.sum(sp[cond]['cs2']) > 0: 
                result['cs1_multi_wo_timecorr'][event_i] = event['s1_area'] * np.average(
                    sp[cond]['s1_xyz_correction_factor'], weights=sp[cond]['cs2'])
                result['cs1_multi'][event_i] = result['cs1_multi_wo_timecorr'][event_i] * np.average(
                    sp[cond]['s1_rel_light_yield_correction_factor'], weights=sp[cond]['cs2'])
                result['x_avg'][event_i] = np.average(sp[cond]['x'], weights=sp[cond]['cs2'])
                result['y_avg'][event_i] = np.average(sp[cond]['y'], weights=sp[cond]['cs2'])
                result['z_obs_avg'][event_i] = np.average(sp[cond]['z_obs_ms'], weights=sp[cond]['cs2'])
                result['cs2_area_fraction_top_avg'][event_i] = np.average(
                    sp[cond]['cs2_area_fraction_top'], weights=sp[cond]['cs2'])   
                result['multiplicity'][event_i] = len(sp[cond]['area'])

        el = self.cs1_to_e(result['cs1_multi'])
        ec = self.cs2_to_e(result['cs2_sum'])
        result['ces_sum'] = el + ec
        result['e_charge_sum'] = ec
        result['time'] = events['time']
        result['endtime'] = strax.endtime(events)
        return result



from enum import IntEnum

import numpy as np
import numba
import strax
import straxen


class PeakSubtyping(IntEnum):
    """Peak Subtyping Dictionary"""
    Undefined = 6
    # Unknown (0)
    Junk = 0
    # S1 (1)
    isoS1 = 10        # S1 without nearby peaks
    S1 = 11           # Regular S1 with a matched pS2
    loneS1 = 12       # Regular S1 without a matched pS2
    sloneS1 = 13      # Regular S1 without a matched pS2, with area < 0.5SE
    # S2 (2)
    DE = 21           # Delayed Extracted few electron peak
    pS2 = 22          # S2 matched with an S1, with area >= 5SE
    isoDE = 23        # S2 without nearby peaks, with area < 5SE
    isopS2 = 24       # S2 without nearby peaks, with area >= 5SE
    lonepS2 = 25      # S2 with area >= 5SE, without a matched S1 but not categorized as other large S2s
    S1olS2 = 261      # S2 with a nearby pS2, with area >= max(0.5 * pS2, 5SE), after S1 but before S2
    S2olS2 = 262      # S2 with a nearby pS2, with area >= max(0.5 * pS2, 5SE), after S2
    S1PH = 271        # Photoionization S2s, with area < max(0.5 * pS2, 5SE), after S1 but before S2
    S2PH = 272        # Photoionization S2s, with area < max(0.5 * pS2, 5SE), after S2
    slS1PH = 273      # Photoionization S2s after a sloneS1
    fakeS2 = 28       # S2 with area < 5SE, with a paired pS2 satisfying S1-S2 time correlation
    fakeS2_olS2 = 29  # olS2 after the fakeS2 and before the associated pS2
    fakeS2_PH = 20    # Photoionization after the fakeS2 and before the associated pS2


class PeaksSubtypes(strax.Plugin):
    """
    Subtyping Peaks
    This plugin scans forward in time and catagorize peaks into subtypes
    based on their correlations with nearby peaks.
    Reference note:
    https://xe1t-wiki.lngs.infn.it/doku.php?id=jlong:peak_subtyping_study
    :returns: an integer index for each peak.
      Please refer to this note:
      https://xe1t-wiki.lngs.infn.it/doku.php?id=jlong:peak_subtyping_dictionary
    """

    __version__ = '0.2.0'
    provides = 'subtype_mask'
    depends_on = ('peak_basics')
    save_when = strax.SaveWhen.NEVER
    dtype = [('subtype', np.int16, 'time subtyping of peaks')] + strax.time_fields

    ref_se_gain = straxen.URLConfig(
        default='bodega://se_gain?bodega_version=v1',
        help='Nominal single electron (SE) gain in PE / electron extracted.')

    ref_se_span = straxen.URLConfig(
        default='bodega://se_spread?bodega_version=v0',
        help = "SE spread value"
    )

    electron_drift_velocity = straxen.URLConfig(
        default='cmt://'
                'electron_drift_velocity'
                '?version=ONLINE&run_id=plugin.run_id',
        cache=True,
        help='Vertical electron drift velocity in cm/ns (1e4 m/ms)'
    )

    max_drift_length = straxen.URLConfig(
        default=straxen.tpc_z, type=(int, float),
        help='Total length of the TPC from the bottom of gate to the '
             'top of cathode wires [cm]',
    )

    ls2_threshold_ne = straxen.URLConfig(
        default=5., type=(int, float),
        help="cutoff between small S2 and large S2, 5e- for now"
    )

    s1_s2_window_fac = straxen.URLConfig(
        default=1.1, type=(int, float),
        help="window to look for pS2 after S1 in multiples of full drift time,"
             "slightly larger to incorporate spread uncertainties of e- drifting"
    )

    other_ls2_fac = straxen.URLConfig(
        default=0.5, type=(int, float),
        help="threshold to consider other large S2s within a window"
    )

    mis_s2_fac = straxen.URLConfig(
        default=0.5, type=(int, float),
        help="threshold to consider potential misclassified SE"
    )

    after_s1_window_ext_fac = straxen.URLConfig(
        default=1, type=(int, float),
        help="extend scanning window after identified large S1s by this much full drift time,"
             "if no S2 is found in s1_s2_window"
    )

    after_s2_window_ext_fac = straxen.URLConfig(
        default=2, type=(int, float),
        help="extend scanning window after identified primary S2 by this much full drift time"
    )

    def setup(self):
        self.drift_time_max = int(self.max_drift_length / self.electron_drift_velocity)
        self.se_gain = self.ref_se_gain['all_tpc']
        self.se_span = self.ref_se_span['all_tpc']
        self.ls2_threshold = self.ls2_threshold_ne * self.se_gain
        self.mis_s2_threshold = self.mis_s2_fac * self.se_gain
        self.ls2_threshold += np.sqrt(self.ls2_threshold_ne) * self.se_span
        self.s1_s2_window = self.s1_s2_window_fac * self.drift_time_max

    def mark_with_s1(self, peaks, mask):
        '''
        Look after each S1s and classify:
        1. mark the largest peak within s1_s2_window as pS2 (exceeding ls2_threshold)
        2. extend window after pS2
        3. mark all other large S2s as olS2. 
        4. mark all other small S2s as photoionization (PH)

        if pS2 identified, extend the window to 2fdt after such pS2 and mark S2s:
        1. other large S2 if an S2 is larger than ls2_threshold and half the pS2 size
        2. photoionization if not olS2
        '''

        undefined_mask = (mask == PeakSubtyping.Undefined)
        dmask = mask[undefined_mask]
        ts = peaks['time'][undefined_mask]  # time
        tps = peaks['type'][undefined_mask]  # type
        pas = peaks['area'][undefined_mask]  # area
        s1s_t = ts[tps == 1]  # s1 time
        s1s_a = pas[tps == 1]  # s1 area
        s2s_t = ts[tps == 2]  # s2 time
        s2s_a = pas[tps == 2]  # s2 area
        s1_mask = dmask[tps == 1]  # define s1 subtypes
        s2_mask = dmask[tps == 2]  # define s2 subtypes

        for i, s1_t in enumerate(s1s_t):
            # locate pS2:
            ps2_cand_cond = (s2s_t > s1_t) & (s2s_t <= s1_t + self.s1_s2_window)
            if ps2_cand_cond.sum():
                # do we have pS2 or a bunch of small S2s
                max_ind = np.nanargmax(s2s_a[ps2_cand_cond])
                max_val = s2s_a[ps2_cand_cond][max_ind]
                max_time = s2s_t[ps2_cand_cond][max_ind]
                if max_val >= self.ls2_threshold:
                    # pS2 located
                    whole_window_cond = (s2s_t > s1_t)
                    whole_window_cond &= (s2s_t <= s2s_t[
                        ps2_cand_cond][max_ind] + self.after_s2_window_ext_fac * self.drift_time_max)
                    s2_mask_tmp = s2_mask[whole_window_cond]
                    small_s2s_a = s2s_a[whole_window_cond] < np.max([
                        self.ls2_threshold, self.other_ls2_fac * max_val])
                    not_pS2 = s2_mask_tmp != PeakSubtyping.pS2
                    within_dt_s2s_t = s2s_t[whole_window_cond] < max_time
                    s2_mask_tmp[
                        ~small_s2s_a & not_pS2 & within_dt_s2s_t] = PeakSubtyping.S1olS2
                    s2_mask_tmp[
                        ~small_s2s_a & not_pS2 & ~within_dt_s2s_t] = PeakSubtyping.S2olS2
                    s2_mask_tmp[
                        small_s2s_a & not_pS2 & within_dt_s2s_t] = PeakSubtyping.S1PH
                    s2_mask_tmp[
                        small_s2s_a & not_pS2 & ~within_dt_s2s_t] = PeakSubtyping.S2PH
                    # assign the pS2 and S1
                    s2_mask_tmp[s2s_t[whole_window_cond] == max_time] = PeakSubtyping.pS2
                    s1_mask[i] = PeakSubtyping.S1
                    s2_mask[whole_window_cond] = s2_mask_tmp
                else:
                    # no pS2, loneS1 spotted
                    whole_window_cond = (s2s_t > s1_t)
                    whole_window_cond &= (s2s_t <= s1_t + self.after_s1_window_ext_fac * self.drift_time_max)
                    if s1s_a[i] < self.mis_s2_threshold:
                        # Potential misclassified SE signal
                        s2_mask[whole_window_cond] = PeakSubtyping.slS1PH
                        s1_mask[i] = PeakSubtyping.sloneS1
                    else:
                        s2_mask[whole_window_cond] = PeakSubtyping.S1PH
                        s1_mask[i] = PeakSubtyping.loneS1
            else:
                if s1s_a[i] < self.mis_s2_threshold:
                    # Potential misclassified SE signal
                    s1_mask[i] = PeakSubtyping.sloneS1
                else:
                    s1_mask[i] = PeakSubtyping.loneS1
        dmask[tps == 1] = s1_mask
        dmask[tps == 2] = s2_mask
        mask[undefined_mask] = dmask

    def mark_s2s(self, peaks, mask):
        '''
        After marking all peaks after S1s, all that's left are S2s.
        One extra occasion is the "fakeS2".
        If a small S2 is identified and a pS2 can be paired up with it, such small S2 is marked "fake S2". 
        '''

        undefined_mask = (mask == PeakSubtyping.Undefined)
        dmask = mask[undefined_mask]
        pt = peaks['time'][undefined_mask]  # time
        pa = peaks['area'][undefined_mask]  # area

        # load only the ones not being classified
        # add a termination number of while loops
        max_iter = len(peaks)
        current_step = 0
        while (dmask == PeakSubtyping.Undefined).sum() > 0 and current_step < max_iter:
            current_step += 1
            start_s2_t = pt[dmask == PeakSubtyping.Undefined][0]
            start_s2_a = pa[dmask == PeakSubtyping.Undefined][0]

            # first mark lonepS2s
            if start_s2_a >= self.ls2_threshold:
                # assign this S2 as lonepS2
                dmask_tmp = dmask[dmask == PeakSubtyping.Undefined]
                dmask_tmp[0] = PeakSubtyping.lonepS2
                dmask[dmask == PeakSubtyping.Undefined] = dmask_tmp
                # mark upto 2 fdt into olS2 or PH
                whole_window_cond = (pt > start_s2_t) & (
                    pt <= start_s2_t + self.after_s2_window_ext_fac * self.drift_time_max)
                mask_tmp = dmask[whole_window_cond]
                small_ps = pa[whole_window_cond] < self.ls2_threshold
                mask_tmp[~small_ps] = PeakSubtyping.S2olS2
                mask_tmp[small_ps] = PeakSubtyping.S2PH
                dmask[whole_window_cond] = mask_tmp
            else:
                # A small S2 might actually be an S1, so here I also mark suspicious couplings
                ps2_cand_cond = (pt > start_s2_t) & (pt <= start_s2_t + self.s1_s2_window)
                if len(pa[ps2_cand_cond]):
                    # do we have pS2 or a bunch of small S2s
                    max_ind = np.nanargmax(pa[ps2_cand_cond])
                    max_val = pa[ps2_cand_cond][max_ind]
                    max_time = pt[ps2_cand_cond][max_ind]
                    if max_val >= self.ls2_threshold:
                        # a suspicious coupling spotted. label this s2 as fakeS2
                        dmask_tmp = dmask[dmask == PeakSubtyping.Undefined]
                        dmask_tmp[0] = PeakSubtyping.fakeS2
                        dmask[dmask == PeakSubtyping.Undefined] = dmask_tmp
                        whole_window_cond = (pt > start_s2_t)
                        whole_window_cond &= (pt <= max_time + self.after_s2_window_ext_fac * self.drift_time_max)
                        mask_tmp = dmask[whole_window_cond]
                        small_pa = pa[whole_window_cond] < np.max([
                            self.ls2_threshold, self.other_ls2_fac * max_val])
                        within_dt_pt = pt[whole_window_cond] < max_time
                        mask_tmp[
                            ~small_pa & within_dt_pt] = PeakSubtyping.fakeS2_olS2
                        mask_tmp[
                            small_pa & within_dt_pt] = PeakSubtyping.fakeS2_PH
                        mask_tmp[
                            ~small_pa & ~within_dt_pt] = PeakSubtyping.S2olS2
                        mask_tmp[
                            small_pa & ~within_dt_pt] = PeakSubtyping.S2PH
                        mask_tmp[pt[whole_window_cond] == max_time] = PeakSubtyping.pS2
                        dmask[whole_window_cond] = mask_tmp
                    else:
                        dmask_tmp = dmask[dmask == PeakSubtyping.Undefined]
                        dmask_tmp[0] = PeakSubtyping.DE
                        dmask[dmask == PeakSubtyping.Undefined] = dmask_tmp
                else:
                    dmask_tmp = dmask[dmask == PeakSubtyping.Undefined]
                    dmask_tmp[0] = PeakSubtyping.DE
                    dmask[dmask == PeakSubtyping.Undefined] = dmask_tmp
        mask[undefined_mask] = dmask

    def compute(self, peaks):
        mask = np.ones(len(peaks)) * PeakSubtyping.Undefined
        mask[peaks['type'] == 0] = PeakSubtyping.Junk

        self.mark_with_s1(peaks, mask)
        self.mark_s2s(peaks, mask)

        n_undefined = (mask == PeakSubtyping.Undefined).sum()
        if n_undefined:
            raise RuntimeError(f'We still find {n_undefined} peaks undefined!')

        return dict(
            time=peaks['time'],
            endtime=strax.endtime(peaks),
            subtype=mask
        )



import strax
import numpy as np
import numba
import straxen


export, __all__ = strax.exporter()


@export
class EventBasics(strax.Plugin):
    """
    Computes the basic properties of the main/alternative S1/S2 within
    an event.

    The main S1 and alternative S1 are given by the largest two S1-Peaks
    within the event.
    The main S2 is given by the largest S2-Peak within the event, while
    alternative S2 is selected as the largest S2 other than main S2
    in the time window [main S1 time, main S1 time + max drift time].
    """
    __version__ = '1.3.3'

    depends_on = ('events',
                  'peak_basics',
                  'peak_positions',
                  'peak_proximity')
    provides = 'event_basics'
    data_kind = 'events'
    loop_over = 'events'

    electron_drift_velocity = straxen.URLConfig(
        default='cmt://'
                'electron_drift_velocity'
                '?version=ONLINE&run_id=plugin.run_id',
        cache=True,
        help='Vertical electron drift velocity in cm/ns (1e4 m/ms)'
    )

    allow_posts2_s1s = straxen.URLConfig(
        default=False, infer_type=False,
        help="Allow S1s past the main S2 to become the main S1 and S2")

    force_main_before_alt = straxen.URLConfig(
        default=False, infer_type=False,
        help="Make the alternate S1 (and likewise S2) the main S1 if "
             "occurs before the main S1.")

    force_alt_s2_in_max_drift_time = straxen.URLConfig(
        default=True, infer_type=False,
        help="Make sure alt_s2 is in max drift time starting from main S1")

    event_s1_min_coincidence = straxen.URLConfig(
        default=2, infer_type=False,
        help="Event level S1 min coincidence. Should be >= s1_min_coincidence "
             "in the peaklet classification")

    max_drift_length = straxen.URLConfig(
        default=straxen.tpc_z, infer_type=False,
        help='Total length of the TPC from the bottom of gate to the '
             'top of cathode wires [cm]', )

    def infer_dtype(self):
        # Basic event properties
        self._set_posrec_save()
        self._set_dtype_requirements()
        dtype = []
        dtype += strax.time_fields
        dtype += [('n_peaks', np.int32,
                   'Number of peaks in the event'),
                  ('drift_time', np.float32,
                   'Drift time between main S1 and S2 in ns'),
                  ('event_number', np.int64,
                   'Event number in this dataset'),
                  ]

        dtype += self._get_si_dtypes(self.peak_properties)

        dtype += [
            (f's1_max_diff', np.int32,
             f'Main S1 largest time difference between apexes of hits [ns]'),
            (f'alt_s1_max_diff', np.int32,
             f'Alternate S1 largest time difference between apexes of hits [ns]'),
            (f's1_min_diff', np.int32,
             f'Main S1 smallest time difference between apexes of hits [ns]'),
            (f'alt_s1_min_diff', np.int32,
             f'Alternate S1 smallest time difference between apexes of hits [ns]'),
        ]

        dtype += [
            (f's2_x', np.float32,
             f'Main S2 reconstructed X position, uncorrected [cm]'),
            (f's2_y', np.float32,
             f'Main S2 reconstructed Y position, uncorrected [cm]'),
            (f'alt_s2_x', np.float32,
             f'Alternate S2 reconstructed X position, uncorrected [cm]'),
            (f'alt_s2_y', np.float32,
             f'Alternate S2 reconstructed Y position, uncorrected [cm]'),
            (f'area_before_main_s2', np.float32,
             f'Sum of areas before Main S2 [PE]'),
            (f'large_s2_before_main_s2', np.float32,
             f'The largest S2 before the Main S2 [PE]')
        ]

        dtype += self._get_posrec_dtypes()
        return dtype

    def _set_dtype_requirements(self):
        """Needs to be run before inferring dtype as it is needed there"""
        # Properties to store for each peak (main and alternate S1 and S2)
        self.peak_properties = (
            ('time', np.int64, 'start time since unix epoch [ns]'),
            ('center_time', np.int64, 'weighted center time since unix epoch [ns]'),
            ('endtime', np.int64, 'end time since unix epoch [ns]'),
            ('area', np.float32, 'area, uncorrected [PE]'),
            ('n_channels', np.int16, 'count of contributing PMTs'),
            ('n_hits', np.int16, 'count of hits contributing at least one sample to the peak'),
            ('n_competing', np.int32, 'number of competing peaks'),
            ('max_pmt', np.int16, 'PMT number which contributes the most PE'),
            ('max_pmt_area', np.float32, 'area in the largest-contributing PMT (PE)'),
            ('range_50p_area', np.float32, 'width, 50% area [ns]'),
            ('range_90p_area', np.float32, 'width, 90% area [ns]'),
            ('rise_time', np.float32, 'time between 10% and 50% area quantiles [ns]'),
            ('area_fraction_top', np.float32, 'fraction of area seen by the top PMT array'),
            ('tight_coincidence', np.int16, 'Channel within tight range of mean'),
            ('n_saturated_channels', np.int16, 'Total number of saturated channels'),
        )

    def setup(self):
        self.drift_time_max = int(self.max_drift_length / self.electron_drift_velocity)

    @staticmethod
    def _get_si_dtypes(peak_properties):
        """Get properties for S1/S2 from peaks directly"""
        si_dtype = []
        for s_i in [1, 2]:
            # Peak indices
            si_dtype += [
                (f's{s_i}_index', np.int32, f'Main S{s_i} peak index in event'),
                (f'alt_s{s_i}_index', np.int32, f'Alternate S{s_i} peak index in event')]

            # Peak properties
            for name, dt, comment in peak_properties:
                si_dtype += [(f's{s_i}_{name}', dt, f'Main S{s_i} {comment}'),
                             (f'alt_s{s_i}_{name}', dt, f'Alternate S{s_i} {comment}')]

            # Drifts and delays
            si_dtype += [
                (f'alt_s{s_i}_interaction_drift_time', np.float32,
                 f'Drift time using alternate S{s_i} [ns]'),
                (f'alt_s{s_i}_delay', np.int32,
                 f'Time between main and alternate S{s_i} [ns]')]
        return si_dtype

    def _set_posrec_save(self):
        """
        parse x_mlp et cetera if needed to get the algorithms used and
        set required class attributes
        """
        posrec_fields = self.deps['peak_positions'].dtype_for('peak_positions').names
        posrec_names = [d.split('_')[-1] for d in posrec_fields if 'x_' in d]

        # Preserve order. "set" is not ordered and dtypes should always be ordered
        self.pos_rec_labels = list(set(posrec_names))
        self.pos_rec_labels.sort()

        self.posrec_save = [(xy + algo)
                            for xy in ['x_', 'y_']
                            for algo in self.pos_rec_labels]

    def _get_posrec_dtypes(self):
        """Get S2 positions for each of the position reconstruction algorithms"""
        posrec_dtpye = []

        for algo in self.pos_rec_labels:
            # S2 positions
            posrec_dtpye += [
                (f's2_x_{algo}', np.float32,
                 f'Main S2 {algo}-reconstructed X position, uncorrected [cm]'),
                (f's2_y_{algo}', np.float32,
                 f'Main S2 {algo}-reconstructed Y position, uncorrected [cm]'),
                (f'alt_s2_x_{algo}', np.float32,
                 f'Alternate S2 {algo}-reconstructed X position, uncorrected [cm]'),
                (f'alt_s2_y_{algo}', np.float32,
                 f'Alternate S2 {algo}-reconstructed Y position, uncorrected [cm]')]

        return posrec_dtpye

    @staticmethod
    def set_nan_defaults(buffer):
        """
        When constructing the dtype, take extra care to set values to
        np.Nan / -1 (for ints) as 0 might have a meaning
        """
        for field in buffer.dtype.names:
            if np.issubdtype(buffer.dtype[field], np.integer):
                buffer[field][:] = -1
            else:
                buffer[field][:] = np.nan

    def compute(self, events, peaks):
        result = np.zeros(len(events), dtype=self.dtype)
        self.set_nan_defaults(result)

        split_peaks = strax.split_by_containment(peaks, events)

        result['time'] = events['time']
        result['endtime'] = events['endtime']
        result['event_number'] = events['event_number']

        self.fill_events(result, events, split_peaks)
        return result

    # If copy_largest_peaks_into_event is ever numbafied, also numbafy this function
    def fill_events(self, result_buffer, events, split_peaks):
        """Loop over the events and peaks within that event"""
        for event_i, _ in enumerate(events):
            peaks_in_event_i = split_peaks[event_i]
            n_peaks = len(peaks_in_event_i)
            result_buffer[event_i]['n_peaks'] = n_peaks

            if not n_peaks:
                raise ValueError(f'No peaks within event?\n{events[event_i]}')

            self.fill_result_i(result_buffer[event_i], peaks_in_event_i)

    def fill_result_i(self, event, peaks):
        """For a single event with the result_buffer"""
        # Consider S2s first, then S1s (to enable allow_posts2_s1s = False)
        # number_of_peaks=0 selects all available s2 and sort by area
        largest_s2s, s2_idx = self.get_largest_sx_peaks(peaks, s_i=2, number_of_peaks=0)

        if not self.allow_posts2_s1s and len(largest_s2s):
            s1_latest_time = largest_s2s[0]['time']
        else:
            s1_latest_time = np.inf

        largest_s1s, s1_idx = self.get_largest_sx_peaks(
            peaks,
            s_i=1,
            s1_before_time=s1_latest_time,
            s1_min_coincidence=self.event_s1_min_coincidence)

        if self.force_alt_s2_in_max_drift_time:
            s2_idx, largest_s2s = self.find_main_alt_s2(largest_s1s,
                                                        s2_idx,
                                                        largest_s2s,
                                                        self.drift_time_max,
                                                        )
        else:
            # Select only the largest two S2s
            largest_s2s, s2_idx = largest_s2s[0:2], s2_idx[0:2]

        if self.force_main_before_alt:
            s2_order = np.argsort(largest_s2s['time'])
            largest_s2s = largest_s2s[s2_order]
            s2_idx = s2_idx[s2_order]

        self.set_sx_index(event, s1_idx, s2_idx)
        self.set_event_properties(event, largest_s1s, largest_s2s, peaks)

        # Loop over S1s and S2s and over main / alt.
        for s_i, largest_s_i in enumerate([largest_s1s, largest_s2s], 1):
            # Largest index 0 -> main sx, 1 -> alt sx
            for largest_index, main_or_alt in enumerate(['s', 'alt_s']):
                peak_properties_to_save = [name for name, _, _ in self.peak_properties]
                if s_i == 1:
                    peak_properties_to_save += ['max_diff', 'min_diff']
                elif s_i == 2:
                    peak_properties_to_save += ['x', 'y']
                    peak_properties_to_save += self.posrec_save
                field_names = [f'{main_or_alt}{s_i}_{name}' for name in peak_properties_to_save]
                self.copy_largest_peaks_into_event(event,
                                                   largest_s_i,
                                                   largest_index,
                                                   field_names,
                                                   peak_properties_to_save)

    @staticmethod
    @numba.njit
    def find_main_alt_s2(largest_s1s, s2_idx, largest_s2s, drift_time_max):
        """Require alt_s2 happens between main S1 and maximum drift time"""
        if len(largest_s1s) > 0 and len(largest_s2s) > 1:
            # If there is a valid s1-s2 pair and has a second s2, then check alt s2 validity
            s2_after_s1 = largest_s2s['center_time'] > largest_s1s[0]['center_time']
            s2_before_max_drift_time = (largest_s2s['center_time']
                                        - largest_s1s[0]['center_time']) < 1.01 * drift_time_max
            mask = s2_after_s1 & s2_before_max_drift_time
            # The selection avoids main_S2
            mask[0] = True
            # Take main and the largest valid alt_S2
            s2_idx, largest_s2s = s2_idx[mask], largest_s2s[mask]
        return s2_idx[:2], largest_s2s[:2]

    @staticmethod
    @numba.njit
    def set_event_properties(result, largest_s1s, largest_s2s, peaks):
        """Get properties like drift time and area before main S2"""
        # Compute drift times only if we have a valid S1-S2 pair
        if len(largest_s1s) > 0 and len(largest_s2s) > 0:
            result['drift_time'] = largest_s2s[0]['center_time'] - largest_s1s[0]['center_time']
            if len(largest_s1s) > 1:
                result['alt_s1_interaction_drift_time'] = largest_s2s[0]['center_time'] - largest_s1s[1]['center_time']
                result['alt_s1_delay'] = largest_s1s[1]['center_time'] - largest_s1s[0]['center_time']
            if len(largest_s2s) > 1:
                result['alt_s2_interaction_drift_time'] = largest_s2s[1]['center_time'] - largest_s1s[0]['center_time']
                result['alt_s2_delay'] = largest_s2s[1]['center_time'] - largest_s2s[0]['center_time']

        # areas before main S2
        if len(largest_s2s):
            peaks_before_ms2 = peaks[peaks['time'] < largest_s2s[0]['time']]
            result['area_before_main_s2'] = np.sum(peaks_before_ms2['area'])

            s2peaks_before_ms2 = peaks_before_ms2[peaks_before_ms2['type'] == 2]
            if len(s2peaks_before_ms2) == 0:
                result['large_s2_before_main_s2'] = 0
            else:
                result['large_s2_before_main_s2'] = np.max(s2peaks_before_ms2['area'])
        return result

    @staticmethod
    # @numba.njit <- works but slows if fill_events is not numbafied
    def get_largest_sx_peaks(peaks,
                             s_i,
                             s1_before_time=np.inf,
                             s1_min_coincidence=0,
                             number_of_peaks=2):
        """Get the largest S1/S2. For S1s allow a min coincidence and max time"""
        # Find all peaks of this type (S1 or S2)
        s_mask = peaks['type'] == s_i
        if s_i == 1:
            s_mask &= peaks['time'] < s1_before_time
            s_mask &= peaks['tight_coincidence'] >= s1_min_coincidence

        selected_peaks = peaks[s_mask]
        s_index = np.arange(len(peaks))[s_mask]
        largest_peaks = np.argsort(selected_peaks['area'])[-number_of_peaks:][::-1]
        return selected_peaks[largest_peaks], s_index[largest_peaks]

    # If only we could numbafy this... Unfortunatly we cannot.
    # Perhaps we could one day consider doing something like strax.copy_to_buffer
    @staticmethod
    def copy_largest_peaks_into_event(result,
                                      largest_s_i,
                                      main_or_alt_index,
                                      result_fields,
                                      peak_fields,
                                      ):
        """
        For one event, write all the peak_fields (e.g. "area") of the peak
        (largest_s_i) into their associated field in the event (e.g. s1_area),
        main_or_alt_index differentiates between main (index 0) and alt (index 1)
        """
        index_not_in_list_of_largest_peaks = main_or_alt_index >= len(largest_s_i)
        if index_not_in_list_of_largest_peaks:
            # There is no such peak. E.g. main_or_alt_index == 1 but largest_s_i = ["Main S1"]
            # Asking for index 1 doesn't work on a len 1 list of peaks.
            return

        for i, ev_field in enumerate(result_fields):
            p_field = peak_fields[i]
            if p_field not in ev_field:
                raise ValueError("Event fields must derive from the peak fields")
            result[ev_field] = largest_s_i[main_or_alt_index][p_field]

    @staticmethod
    # @numba.njit <- works but slows if fill_events is not numbafied
    def set_sx_index(res, s1_idx, s2_idx):
        if len(s1_idx):
            res['s1_index'] = s1_idx[0]
            if len(s1_idx) > 1:
                res['alt_s1_index'] = s1_idx[1]
        if len(s2_idx):
            res['s2_index'] = s2_idx[0]
            if len(s2_idx) > 1:
                res['alt_s2_index'] = s2_idx[1]



import strax
import numpy as np
import straxen

export, __all__ = strax.exporter()


@export
class Events(strax.OverlapWindowPlugin):
    """
    Plugin which defines an "event" in our TPC.

    An event is defined by peak(s) in fixed range of time around a peak
    which satisfies certain conditions:
        1. The triggering peak must have a certain area.
        2. The triggering peak must have less than
           "trigger_max_competing" peaks. (A competing peak must have a
           certain area fraction of the triggering peak and must be in a
           window close to the main peak)

    Note:
        The time range which defines an event gets chopped at the chunk
        boundaries. This happens at invalid boundaries of the
    """

    __version__ = '0.1.1'

    depends_on = ['peak_basics', 'peak_proximity']
    provides = 'events'
    data_kind = 'events'

    save_when = strax.SaveWhen.NEVER

    dtype = [
        ('event_number', np.int64, 'Event number in this dataset'),
        ('time', np.int64, 'Event start time in ns since the unix epoch'),
        ('endtime', np.int64, 'Event end time in ns since the unix epoch')]

    events_seen = 0

    electron_drift_velocity = straxen.URLConfig(
        default='cmt://'
                'electron_drift_velocity'
                '?version=ONLINE&run_id=plugin.run_id',
        cache=True,
        help='Vertical electron drift velocity in cm/ns (1e4 m/ms)'
    )

    trigger_min_area = straxen.URLConfig(
        default=100, type=(int, float),
        help='Peaks must have more area (PE) than this to '
             'cause events')

    trigger_max_competing = straxen.URLConfig(
        default=7, type=int,
        help='Peaks must have FEWER nearby larger or slightly smaller'
             ' peaks to cause events')

    left_event_extension = straxen.URLConfig(
        default=int(0.25e6), type=(int, float),
        help='Extend events this many ns to the left from each '
             'triggering peak. This extension is added to the maximum '
             'drift time.',
    )

    right_event_extension = straxen.URLConfig(
        default=int(0.25e6), type=(int, float),
        help='Extend events this many ns to the right from each '
             'triggering peak.',
    )

    max_drift_length = straxen.URLConfig(
        default=straxen.tpc_z, type=(int, float),
        help='Total length of the TPC from the bottom of gate to the '
             'top of cathode wires [cm]',
    )

    exclude_s1_as_triggering_peaks = straxen.URLConfig(
        default=True, type=bool,
        help='If true exclude S1s as triggering peaks.',
    )

    event_s1_min_coincidence = straxen.URLConfig(
        default=2, infer_type=False,
        help="Event level S1 min coincidence. Should be >= "
             "s1_min_coincidence in the peaklet classification")

    s1_min_coincidence = straxen.URLConfig(
        default=2, type=int,
        help="Minimum tight coincidence necessary to make an S1")

    diagnose_overlapping = straxen.URLConfig(
        track=False, default=True, infer_type=False,
        help="Enable runtime checks for disjointness")

    def setup(self):
        if self.s1_min_coincidence > self.event_s1_min_coincidence:
            raise ValueError('Peak s1 coincidence requirement should be smaller '
                             'or equal to event_s1_min_coincidence')
        self.drift_time_max = int(self.max_drift_length / self.electron_drift_velocity)
        # Left_extension and right_extension should be computed in setup to be
        # reflected in cutax too.
        self.left_extension = self.left_event_extension + self.drift_time_max
        self.right_extension = self.right_event_extension

    def get_window_size(self):
        # Take a large window for safety, events can have long tails
        return 10 * (self.left_event_extension
                     + self.drift_time_max
                     + self.right_event_extension)

    def compute(self, peaks, start, end):
        _is_triggering = peaks['area'] > self.trigger_min_area
        _is_triggering &= (peaks['n_competing'] <= self.trigger_max_competing)
        if self.exclude_s1_as_triggering_peaks:
            _is_triggering &= peaks['type'] == 2
        else:
            is_not_s1 = peaks['type'] != 1
            has_tc_large_enough = (peaks['tight_coincidence']
                                   >= self.event_s1_min_coincidence)
            _is_triggering &= (is_not_s1 | has_tc_large_enough)

        triggers = peaks[_is_triggering]

        # Join nearby triggers
        t0, t1 = strax.find_peak_groups(
            triggers,
            gap_threshold=self.left_extension + self.right_extension + 1,
            left_extension=self.left_extension,
            right_extension=self.right_extension)

        # Don't extend beyond the chunk boundaries
        # This will often happen for events near the invalid boundary of the
        # overlap processing (which should be thrown away)
        t0 = np.clip(t0, start, end)
        t1 = np.clip(t1, start, end)

        result = np.zeros(len(t0), self.dtype)
        result['time'] = t0
        result['endtime'] = t1
        result['event_number'] = np.arange(len(result)) + self.events_seen

        if not result.size > 0:
            print("Found chunk without events?!")

        if self.diagnose_overlapping and len(result):
            # Check if the event windows overlap
            _event_window_do_not_overlap = (strax.endtime(result)[:-1] - result['time'][1:]) <= 0
            assert np.all(_event_window_do_not_overlap), "Events not disjoint"

        self.events_seen += len(result)

        return result



"""
Core functions of the DAQ, mostly used in straxen/bin
"""

import pymongo
import utilix
from utilix import uconfig
import straxen
from datetime import datetime, timedelta
import pytz

ceph_folder = '/live_data/xenonnt/'
output_folder = '/data/xenonnt_processed/'
pre_folder = '/data/pre_processed/'
non_registered_folder = '/data/xenonnt_unregistered/'


class DataBases:
    def __init__(self, production=False):
        self.production = production
        # DAQ database
        daq_db_name = 'daq'
        daq_uri = straxen.get_mongo_uri(header='rundb_admin',
                                        user_key='mongo_daq_username',
                                        pwd_key='mongo_daq_password',
                                        url_key='mongo_daq_url')
        daq_client = pymongo.MongoClient(daq_uri)
        self.daq_db = daq_client[daq_db_name]
        self.bs_coll = self.daq_db['eb_monitor']
        self.ag_stat_coll = self.daq_db['aggregate_status']
        self.log_coll = self.daq_db['log']

        # Runs database
        run_dbname = straxen.uconfig.get('rundb_admin', 'mongo_rdb_database')
        run_collname = 'runs'
        if production:
            self.run_db = self.get_admin_client()[run_dbname]
        else:
            # Please note, this is a read only account on the rundb
            run_uri = straxen.get_mongo_uri()
            run_client = pymongo.MongoClient(run_uri)
            self.run_db = run_client[run_dbname]
        self.run_coll = self.run_db[run_collname]

    @staticmethod
    def get_admin_client():
        # We want admin access to start writing data!
        mongo_url = uconfig.get('rundb_admin', 'mongo_rdb_url')
        mongo_user = uconfig.get('rundb_admin', 'mongo_rdb_username')
        mongo_password = uconfig.get('rundb_admin', 'mongo_rdb_password')
        mongo_database = uconfig.get('rundb_admin', 'mongo_rdb_database')

        collection = utilix.rundb.xent_collection(
            url=mongo_url, user=mongo_user, password=mongo_password, database=mongo_database)

        # Do not delete the client!
        return collection.database.client

    def log_warning(self,
                    message,
                    priority='warning',
                    run_id=None,
                    production=True,
                    user='daq_process',
                    ):
        """Report a warning to the terminal (using the logging module)
        and the DAQ log DB.
        :param message: insert string into log_coll
        :param priority: severity of warning. Can be:
            info: 1,
            warning: 2,
            <any other valid python logging level, e.g. error or fatal>: 3
        :param run_id: optional run id.
        """
        if not production:
            return

        # Log according to redax rules
        # https://github.com/coderdj/redax/blob/master/MongoLog.hh#L22
        warning_message = {
            'message': message,
            'user': user,
            'priority': dict(debug=0, info=1, warning=2, error=3, fatal=4, ).get(priority.lower(), 3)}
        if run_id is not None:
            warning_message.update({'runid': int(run_id)})
        self.log_coll.insert_one(warning_message)


def now(plus=0):
    """Now in utc time"""
    return datetime.now(pytz.utc) + timedelta(seconds=plus)



"""S1 Position reconstruction for Xenon-nT
Author and maintainer: Matteo Guida (guidam@mpi-hd.mpg.de)"""

import strax
import straxen
import numpy as np
from warnings import warn
import time
export, __all__ = strax.exporter()

@export
class PeakS1PositionBase(strax.Plugin):
    """
    Base pluging for S1 position reconstruction at peak level.
    Provides x_algorithm, y_algorithm, z_algorithm for all peaks > than
    min-reconstruction area based on both the top and bottom array.
    """
    __version__ = '0.0.0'
    depends_on = ('peaks',)

    algorithm = None
    compressor = 'zstd'
    parallel = True  # can set to "process" after #82

    min_s1_area_s1_posrec = straxen.URLConfig(
        help='Skip reconstruction if area (PE) is less than this',
        default=1000, infer_type=False)

    def infer_dtype(self):
        if self.algorithm is None:
            raise NotImplementedError(f'Base class should not be used without '
                                      f'algorithm as done in {__class__.__name__}')

    def infer_dtype(self):
        if self.algorithm is None:
            raise NotImplementedError(f'Base class should not be used without '
                                      f'algorithm as done in {__class__.__name__}')
        dtype = [('x_' + self.algorithm, np.float32,
                  f'Reconstructed {self.algorithm} S1 X position (cm), uncorrected'),
                 ('y_' + self.algorithm, np.float32,
                  f'Reconstructed {self.algorithm} S1 Y position (cm), uncorrected'),
                 ('z_' + self.algorithm, np.float32,
                  f'Reconstructed {self.algorithm} S1 Z position (cm), uncorrected')
                  ]

        dtype += strax.time_fields
        return dtype
    
    def get_tf_model(self):
        """
        Simple wrapper to have several tf_model_s1_cnn, ..
        point to this same function in the compute method
        """
        model = getattr(self, f'tf_model_{self.algorithm}', None)
        if model is None:
            warn(f'Setting model to None for {self.__class__.__name__} will '
                 f'set only nans as output for {self.algorithm}')
        if isinstance(model, str):
            raise ValueError(f'open files from tf:// protocol! Got {model} '
                             f'instead, see tests/test_s1_posrec.py for examples.')
        return model

    def compute(self, peaks):
        result = np.ones(len(peaks), dtype=self.dtype)
        result['time'], result['endtime'] = peaks['time'], strax.endtime(peaks)

        result['x_' + self.algorithm] *= float('nan')
        result['y_' + self.algorithm] *= float('nan')
        result['z_' + self.algorithm] *= float('nan')

        model = self.get_tf_model()

        # Reconstruct position only for large peaks, otherwise severe inaccuracy.
        peak_mask = peaks['area_per_channel'].sum(axis=1) > self.min_s1_area_s1_posrec
        peak_mask &= peaks['type'] < 2

        if not np.sum(peak_mask):
            # No peaks fulfilling the conditions, return nan array.
            return result

        _in = peaks['area_per_channel'][peak_mask]

        with np.errstate(divide='ignore', invalid='ignore'):
        # Normalise patters by dividing by largest PMT output between the two arrays. 
            _in = _in / _in.max(axis=1, keepdims=True)

        # Getting actual position reconstruction
        _out = model.predict(_in)

        # writing output to the result
        result['x_' + self.algorithm][peak_mask] = _out[:, 0]
        result['y_' + self.algorithm][peak_mask] = _out[:, 1]
        result['z_' + self.algorithm][peak_mask] = _out[:, 2]    
        return result



import warnings

import numba
import numpy as np
import strax
import straxen

DEFAULT_MAX_SAMPLES = 20_000


@straxen.mini_analysis(requires=('records',),
                       warn_beyond_sec=10,
                       default_time_selection='touching')
def records_matrix(records, time_range, seconds_range, config, to_pe,
                   max_samples=DEFAULT_MAX_SAMPLES,
                   ignore_max_sample_warning=False):
    """Return (wv_matrix, times, pms)
      - wv_matrix: (n_samples, n_pmt) array with per-PMT waveform intensity in PE/ns
      - times: time labels in seconds (corr. to rows)
      - pmts: PMT numbers (corr. to columns)
    Both times and pmts have one extra element.

    :param max_samples: Maximum number of time samples. If window and dt
    conspire to exceed this, waveforms will be downsampled.
    :param ignore_max_sample_warning: If True, suppress warning when this happens.

    Example:
        wvm, ts, ys = st.records_matrix(run_id, seconds_range=(1., 1.00001))
        plt.pcolormesh(ts, ys, wvm.T,
                       norm=matplotlib.colors.LogNorm())
        plt.colorbar(label='Intensity [PE / ns]')
    """
    if len(records):
        dt = records[0]['dt']
        samples_per_record = len(records[0]['data'])
    else:
        # Defaults here do not matter, nothing will be plotted anyway
        dt = 10, 110
    record_duration = samples_per_record * dt

    window = time_range[1] - time_range[0]
    if window / dt > max_samples:
        with np.errstate(divide='ignore', invalid='ignore'):
            # Downsample. New dt must be
            #  a) multiple of old dt
            dts = np.arange(0, record_duration + dt, dt).astype(np.int64)
            #  b) divisor of record duration
            dts = dts[record_duration / dts % 1 == 0]
            #  c) total samples < max_samples
            dts = dts[window / dts < max_samples]
            if len(dts):
                # Pick lowest dt that satisfies criteria
                dt = dts.min()
            else:
                # Records will be downsampled to single points
                dt = max(record_duration, window // max_samples)
        if not ignore_max_sample_warning:
            warnings.warn(f"Matrix would exceed max_samples {max_samples}, "
                          f"downsampling to dt = {dt} ns.")

    wvm = _records_to_matrix(
        records,
        t0=time_range[0],
        n_channels=config['n_tpc_pmts'],
        dt=dt,
        window=window)
    wvm = wvm.astype(np.float32) * to_pe.reshape(1, -1) / dt

    # Note + 1, so data for sample 0 will range from 0-1 in plot
    ts = (np.arange(wvm.shape[0] + 1) * dt / int(1e9) + seconds_range[0])
    ys = np.arange(wvm.shape[1] + 1)

    return wvm, ts, ys


@straxen.mini_analysis(requires=('raw_records',),
                       warn_beyond_sec=3e-3,
                       default_time_selection='touching')
def raw_records_matrix(context, run_id, raw_records, time_range,
                       ignore_max_sample_warning=False,
                       max_samples=DEFAULT_MAX_SAMPLES,
                       **kwargs):
    # Convert raw to records. We may not be able to baseline correctly
    # at the start of the range due to missing zeroth fragments
    records = strax.raw_to_records(raw_records)
    strax.baseline(records, allow_sloppy_chunking=True)
    strax.zero_out_of_bounds(records)

    return context.records_matrix(run_id=run_id,
                                  records=records,
                                  time_range=time_range,
                                  max_samples=max_samples,
                                  ignore_max_sample_warning=ignore_max_sample_warning,
                                  **kwargs)


def _records_to_matrix(records, t0, window, n_channels, dt=10):
    if np.any(records['amplitude_bit_shift'] > 0):
        warnings.warn('Ignoring amplitude bitshift!')
    return _records_to_matrix_inner(records, t0, window, n_channels, dt)


@numba.njit
def _records_to_matrix_inner(records, t0, window, n_channels, dt=10):
    n_samples = (window // dt) + 1
    # Use 32-bit integers, so downsampling saturated samples doesn't
    # cause wraparounds
    y = np.zeros((n_samples, n_channels),
                 dtype=np.int32)

    if not len(records):
        return y
    samples_per_record = len(records[0]['data'])

    for r in records:
        if r['channel'] > n_channels:
            continue

        if dt >= samples_per_record * r['dt']:
            # Downsample to single sample -> store area
            idx = (r['time'] - t0) // dt
            if idx >= len(y):
                print(len(y), idx)
                raise IndexError('Despite n_samples = window // dt + 1, our '
                                 'idx is too high?!')
            y[idx, r['channel']] += r['area']
            continue

        # Assume out-of-bounds data has been zeroed, so we do not
        # need to do r['data'][:r['length']] here.
        # This simplifies downsampling.
        w = r['data'].astype(np.int32)

        if dt > r['dt']:
            # Downsample
            duration = samples_per_record * r['dt']
            if duration % dt != 0:
                raise ValueError("Cannot downsample fractionally")
            # .astype here keeps numba happy ... ??
            w = w.reshape(duration // dt, -1).sum(axis=1).astype(np.int32)

        elif dt < r['dt']:
            raise ValueError("Upsampling not yet implemented")

        (r_start, r_end), (y_start, y_end) = strax.overlap_indices(
            r['time'] // dt, len(w),
            t0 // dt, n_samples)
        # += is paranoid, data in individual channels should not overlap
        # but... https://github.com/AxFoundation/strax/issues/119
        y[y_start:y_end, r['channel']] += w[r_start:r_end]

    return y



import strax
import straxen
from straxen.plugins.events._event_s2_positions_base import EventS2PositionBase


export, __all__ = strax.exporter()


@export
class EventS2PositionGCN(EventS2PositionBase):
    """
    GCN net for position S2 reconstruction at event level
    """
    algorithm = "gcn"
    provides = "event_s2_positions_gcn"

    tf_event_model_gcn = straxen.URLConfig(
        default=f'tf://'
                f'resource://'
                f'cmt://{algorithm}_model'
                f'?version=ONLINE'
                f'&run_id=plugin.run_id'
                f'&fmt=abs_path',
        help='MLP model. Should be opened using the "tf" descriptor. '
             'Set to "None" to skip computation',
        cache=3,
    )



import numpy as np
import strax
import straxen
from straxen.plugins.hitlets_nv.hitlets_nv import nVETOHitlets
from straxen.plugins.defaults import MV_PREAMBLE, NV_HIT_DEFAULTS, MV_HIT_DEFAULTS

export, __all__ = strax.exporter()


@export
class muVETOHitlets(nVETOHitlets):
    __doc__ = MV_PREAMBLE + nVETOHitlets.__doc__
    __version__ = '0.0.2'
    depends_on = 'records_mv'

    provides = 'hitlets_mv'
    data_kind = 'hitlets_mv'
    child_plugin = True

    dtype = strax.hitlet_dtype()

    save_outside_hits_mv = straxen.URLConfig(
        default=MV_HIT_DEFAULTS['save_outside_hits_mv'], track=True, infer_type=False,
        child_option=True, parent_option_name='save_outside_hits_nv',
        help='Save (left, right) samples besides hits; cut the rest')

    hit_min_amplitude_mv = straxen.URLConfig(
        infer_type=False,
        default=MV_HIT_DEFAULTS['hit_min_amplitude_mv'], track=True,
        help='Minimum hit amplitude in ADC counts above baseline. '
             'Specify as a tuple of length n_mveto_pmts, or a number, '
             'or a string like "pmt_commissioning_initial" which means calling '
             'hitfinder_thresholds.py, '
             'or a tuple like (correction=str, version=str, nT=boolean),'
             'which means we are using cmt.')

    min_split_mv = straxen.URLConfig(
        default=100, track=True, infer_type=False,
        child_option=True, parent_option_name='min_split_nv',
        help='Minimum height difference pe/sample between local minimum and maximum, '
             'that a pulse get split.')

    min_split_ratio_mv = straxen.URLConfig(
        default=0, track=True, infer_type=False,
        child_option=True, parent_option_name='min_split_ratio_nv',
        help='Min ratio between local maximum and minimum to split pulse (zero to switch this '
             'off).')

    entropy_template_mv = straxen.URLConfig(
        default='flat', track=True, infer_type=False,
        child_option=True, parent_option_name='entropy_template_nv',
        help='Template data is compared with in conditional entropy. Can be either "flat" or a '
             'template array.')

    entropy_square_data_mv = straxen.URLConfig(
        default=False, track=True, infer_type=False,
        child_option=True, parent_option_name='entropy_square_data_nv',
        help='Parameter which decides if data is first squared before normalized and compared to '
             'the template.')

    gain_model_mv = straxen.URLConfig(
        default="cmt://to_pe_model_mv?version=ONLINE&run_id=plugin.run_id", infer_type=False,
        child_option=True, parent_option_name='gain_model_nv',
        help='PMT gain model. Specify as (model_type, model_config)')

    def setup(self):
        self.channel_range = self.channel_map['mv']
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1

        to_pe = self.gain_model_mv
        # Create to_pe array of size max channel:
        self.to_pe = np.zeros(self.channel_range[1] + 1, dtype=np.float32)
        self.to_pe[self.channel_range[0]:] = to_pe[:]

        self.hit_thresholds = self.hit_min_amplitude_mv

    def compute(self, records_mv, start, end):
        return super().compute(records_mv, start, end)



import numpy as np
import strax
import straxen


export, __all__ = strax.exporter()


@export
class PeakletClassification(strax.Plugin):
    """Classify peaklets as unknown, S1, or S2."""
    __version__ = '3.0.3'

    provides = 'peaklet_classification'
    depends_on = ('peaklets',)
    parallel = True
    dtype = (strax.peak_interval_dtype
             + [('type', np.int8, 'Classification of the peak(let)')])

    s1_risetime_area_parameters = straxen.URLConfig(
        default=(50, 80, 12), type=(list, tuple),
        help="norm, const, tau in the empirical boundary in the risetime-area plot")

    s1_risetime_aft_parameters = straxen.URLConfig(
        default=(-1, 2.6), type=(list, tuple),
        help=(
            "Slope and offset in exponential of emperical boundary in the rise time-AFT "
            "plot. Specified as (slope, offset)"))

    s1_flatten_threshold_aft = straxen.URLConfig(
        default=(0.6, 100), type=(tuple, list),
        help=(
            "Threshold for AFT, above which we use a flatted boundary for rise time"
            "Specified values: (AFT boundary, constant rise time)."))

    n_top_pmts = straxen.URLConfig(
        default=straxen.n_top_pmts, type=int,
        help="Number of top PMTs")

    s1_max_rise_time_post100 = straxen.URLConfig(
        default=200, type=(int, float),
        help="Maximum S1 rise time for > 100 PE [ns]")

    s1_min_coincidence = straxen.URLConfig(
        default=2, type=int,
        help="Minimum tight coincidence necessary to make an S1")

    s2_min_pmts = straxen.URLConfig(
        default=4, type=int,
        help="Minimum number of PMTs contributing to an S2")

    @staticmethod
    def upper_rise_time_area_boundary(area, norm, const, tau):
        """
        Function which determines the upper boundary for the rise-time
        for a given area.
        """
        return norm * np.exp(-area / tau) + const

    @staticmethod
    def upper_rise_time_aft_boundary(aft, slope, offset, aft_boundary, flat_threshold):
        """
        Function which computes the upper rise time boundary as a function
        of area fraction top.
        """
        res = 10 ** (slope * aft + offset)
        res[aft >= aft_boundary] = flat_threshold
        return res

    def compute(self, peaklets):
        ptype = np.zeros(len(peaklets), dtype=np.int8)

        # Properties needed for classification:
        rise_time = -peaklets['area_decile_from_midpoint'][:, 1]
        n_channels = (peaklets['area_per_channel'] > 0).sum(axis=1)
        n_top = self.n_top_pmts
        area_top = peaklets['area_per_channel'][:, :n_top].sum(axis=1)
        area_total = peaklets['area_per_channel'].sum(axis=1)
        area_fraction_top = area_top / area_total

        is_large_s1 = (peaklets['area'] >= 100)
        is_large_s1 &= (rise_time <= self.s1_max_rise_time_post100)
        is_large_s1 &= peaklets['tight_coincidence'] >= self.s1_min_coincidence

        is_small_s1 = peaklets["area"] < 100
        is_small_s1 &= rise_time < self.upper_rise_time_area_boundary(
            peaklets["area"],
            *self.s1_risetime_area_parameters,
        )

        is_small_s1 &= rise_time < self.upper_rise_time_aft_boundary(
            area_fraction_top,
            *self.s1_risetime_aft_parameters,
            *self.s1_flatten_threshold_aft,
        )

        is_small_s1 &= peaklets['tight_coincidence'] >= self.s1_min_coincidence

        ptype[is_large_s1 | is_small_s1] = 1

        is_s2 = n_channels >= self.s2_min_pmts
        is_s2[is_large_s1 | is_small_s1] = False
        ptype[is_s2] = 2

        return dict(type=ptype,
                    time=peaklets['time'],
                    dt=peaklets['dt'],
                    # Channel is added so the field order of the merger of
                    # peaklet_classification and peaklets matches that
                    # of peaklets.
                    # This way S2 merging works on arrays of the same dtype.
                    channel=-1,
                    length=peaklets['length'])



import strax
import numpy as np
import straxen

export, __all__ = strax.exporter()


@export
class OnlinePeakMonitor(strax.Plugin):
    """
    Plugin to write data to the online-monitor. Data that is written by
    this plugin should be small such as to not overload the runs-
    database.

    This plugin takes 'peak_basics' and 'lone_hits'. Although they are
    not strictly related, they are aggregated into a single data_type
    in order to minimize the number of documents in the online monitor.

    Produces 'online_peak_monitor' with info on the lone-hits and peaks
    """

    __version__ = '0.0.5'

    depends_on = ('peak_basics', 'lone_hits')
    provides = 'online_peak_monitor'
    data_kind = 'online_peak_monitor'

    area_vs_width_nbins = straxen.URLConfig(
        type=int, default=60,
        help='Number of bins for area vs width histogram for online monitor. '
             'NB: this is a 2D histogram')

    area_vs_width_bounds = straxen.URLConfig(
        type=tuple, default=((0, 5), (0, 5)),
        help='Boundaries of log-log histogram of area vs width')

    area_vs_width_cut_string = straxen.URLConfig(
        type=str, default='',
        help='Selection (like selection_str) applied to data for '
             '"area_vs_width_hist_clean", cuts should be separated using "&"'
             'For example: (tight_coincidence > 2) & (area_fraction_top < 0.1)'
             'Default is no selection (other than "area_vs_width_min_gap")')

    lone_hits_area_bounds = straxen.URLConfig(
        type=tuple, default=(0, 1500),
        help='Boundaries area histogram of lone hits [ADC]')

    online_peak_monitor_nbins = straxen.URLConfig(
        type=int, default=100,
        help='Number of bins of histogram of online monitor. Will be used '
             'for: '
             'lone_hits_area-histogram, '
             'area_fraction_top-histogram, '
             'online_se_gain estimate (histogram is not stored), '
    )

    lone_hits_cut_string = straxen.URLConfig(
        type=str,
        default='(area >= 50) & (area <= 250)',
        help='Selection (like selection_str) applied to data for '
             '"lone-hits", cuts should be separated using "&")')

    lone_hits_min_gap = straxen.URLConfig(
        type=int,
        default=15_000,
        help='Minimal gap [ns] between consecutive lone-hits. To turn off '
             'this cut, set to 0.')

    n_tpc_pmts = straxen.URLConfig(
        type=int,
        help='Number of TPC PMTs')

    online_se_bounds = straxen.URLConfig(
        type=tuple, default=(7, 70),
        help='Window for online monitor [PE] to look for the SE gain, value'
    )

    def infer_dtype(self):
        n_bins_area_width = self.area_vs_width_nbins
        bounds_area_width = self.area_vs_width_bounds

        n_bins = self.online_peak_monitor_nbins

        n_tpc_pmts = self.n_tpc_pmts
        dtype = [
            (('Start time of the chunk', 'time'),
             np.int64),
            (('End time of the chunk', 'endtime'),
             np.int64),
            (('Area vs width histogram (log-log)', 'area_vs_width_hist'),
             (np.int64, (n_bins_area_width, n_bins_area_width))),
            (('Area vs width edges (log-space)', 'area_vs_width_bounds'),
             (np.float64, np.shape(bounds_area_width))),
            (('Lone hits areas histogram [ADC-counts]', 'lone_hits_area_hist'),
             (np.int64, n_bins)),
            (('Lone hits areas bounds [ADC-counts]', 'lone_hits_area_bounds'),
             (np.float64, 2)),
            (('Lone hits per channel', 'lone_hits_per_channel'),
             (np.int64, n_tpc_pmts)),
            (('AFT histogram', 'aft_hist'),
             (np.int64, n_bins)),
            (('AFT bounds', 'aft_bounds'),
             (np.float64, 2)),
            (('Number of contributing channels histogram', 'n_channel_hist'),
             (np.int64, n_tpc_pmts)),
            (('Single electron gain', 'online_se_gain'),
             np.float32),
        ]
        return dtype

    def compute(self, peaks, lone_hits, start, end):
        # General setup
        res = np.zeros(1, dtype=self.dtype)
        res['time'] = start
        res['endtime'] = end
        n_pmt = self.n_tpc_pmts
        n_bins = self.online_peak_monitor_nbins

        # Bounds for histograms
        res['area_vs_width_bounds'] = self.area_vs_width_bounds
        res['lone_hits_area_bounds'] = self.lone_hits_area_bounds

        # -- Peak vs area 2D histogram --
        # Always cut out unphysical peaks
        sel = (peaks['area'] > 0) & (peaks['range_50p_area'] > 0)
        res['area_vs_width_hist'] = self.area_width_hist(peaks[sel])
        del sel

        # -- Lone hit properties --
        # Make a mask with the cuts.
        # Now only take lone hits that are separated in time.
        if len(lone_hits):
            lh_timedelta = lone_hits[1:]['time'] - strax.endtime(lone_hits)[:-1]
            # Hits on the left are far away? (assume first is because of chunk bound)
            mask = np.hstack([True, lh_timedelta > self.lone_hits_min_gap])
            # Hits on the right are far away? (assume last is because of chunk bound)
            mask &= np.hstack([lh_timedelta > self.lone_hits_min_gap, True])
        else:
            mask = []
        masked_lh = strax.apply_selection(lone_hits[mask],
                                          selection_str=self.lone_hits_cut_string)

        # Make histogram of ADC counts
        # NB: LONE HITS AREA ARE IN ADC!
        lone_hit_areas, _ = np.histogram(masked_lh['area'],
                                         bins=n_bins,
                                         range=self.lone_hits_area_bounds)

        lone_hit_channel_count, _ = np.histogram(masked_lh['channel'],
                                                 bins=n_pmt,
                                                 range=[0, n_pmt])
        # Count number of lone-hits per PMT
        res['lone_hits_area_hist'] = lone_hit_areas
        res['lone_hits_per_channel'] = lone_hit_channel_count
        # Clear mask, don't re-use
        del mask

        # -- AFT histogram --
        aft_b = [0, 1]
        aft_hist, _ = np.histogram(peaks['area_fraction_top'], bins=n_bins, range=aft_b)
        res['aft_hist'] = aft_hist
        res['aft_bounds'] = aft_b

        # Estimate Single Electron (SE) gain
        se_hist, se_bins = np.histogram(peaks['area'], bins=n_bins,
                                        range=self.online_se_bounds)
        bin_centers = (se_bins[1:] + se_bins[:-1]) / 2
        res['online_se_gain'] = bin_centers[np.argmax(se_hist)]
        return res

    def area_width_hist(self, data):
        """Make area vs width 2D-hist"""
        hist, _, _ = np.histogram2d(
            np.log10(data['area']),
            np.log10(data['range_50p_area']),
            range=self.area_vs_width_bounds,
            bins=self.area_vs_width_nbins)
        return hist.T



"""
Test certain interactions with the runsdatabase.
NB! this only works if one has access to the database. This does not
work e.g. on travis jobs and therefore the tests failing locally will
not show up in Pull Requests.
"""
import straxen
import os
import unittest


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
class TestSelectRuns(unittest.TestCase):
    def test_select_runs(self, check_n_runs=2):
        """
        Test (if we have a connection) if we can perform strax.select_runs
            on the last two runs in the runs collection

        :param check_n_runs: int, the number of runs we want to check
        """
        self.assertTrue(check_n_runs >= 1)
        st = straxen.contexts.xenonnt_online(use_rucio=False)
        run_col = st.storage[0].collection

        # Find the latest run in the runs collection
        last_run = run_col.find_one(projection={'number': 1},
                                    sort=[('number', -1)]
                                    ).get('number')

        # Set this number as the minimum run number. This limits the
        # amount of documents checked and therefore keeps the test short.
        st.storage[0].minimum_run_number = int(last_run) - (check_n_runs - 1)
        st.select_runs()


@unittest.skipIf(not straxen.utilix_is_configured(),
                 "Cannot download because utilix is not configured")
class TestDownloader(unittest.TestCase):
    def test_downloader(self):
        """Test if we can download a small file from the downloader"""
        downloader = straxen.MongoDownloader()
        path = downloader.download_single('to_pe_nt.npy')
        self.assertTrue(os.path.exists(path))



from . import bokeh_waveform_plot
from . import daq_waveforms
from . import event_display
from . import holoviews_waveform_display
from . import posrec_comparison
from . import pulse_plots
from . import quick_checks
from . import records_matrix
from . import waveform_plot



import numpy as np
from immutabledict import immutabledict

import strax
import straxen

from straxen.plugins.defaults import MV_PREAMBLE, NV_HIT_DEFAULTS, MV_HIT_DEFAULTS

export, __all__ = strax.exporter()


@export
class nVETOHitlets(strax.Plugin):
    """
    Plugin which computes the nveto hitlets and their parameters.

    Hitlets are an extension of regular hits. They include the left
    and right extension. The plugin does the following:
        1. Generate hitlets which includes these sub-steps:
            * Apply left and right hit extension and concatenate
            overlapping hits.
            * Generate temp. hitelts and look for their waveforms in
            their corresponding records.
            * Split hitlets if they satisfy the set criteria.
        2. Compute the properties of the hitlets.

    Note:
        Hitlets are getting chopped if extended in not recorded regions.
    """
    __version__ = '0.1.1'

    parallel = 'process'
    rechunk_on_save = True
    compressor = 'zstd'

    depends_on = 'records_nv'

    provides = 'hitlets_nv'
    data_kind = 'hitlets_nv'

    dtype = strax.hitlet_dtype()

    save_outside_hits_nv = straxen.URLConfig(
        default=NV_HIT_DEFAULTS['save_outside_hits_nv'], track=True, infer_type=False,
        help='Save (left, right) samples besides hits; cut the rest')

    hit_min_amplitude_nv = straxen.URLConfig(
        infer_type=False,
        default=NV_HIT_DEFAULTS['hit_min_amplitude_nv'], track=True,
        help='Minimum hit amplitude in ADC counts above baseline. '
             'Specify as a tuple of length n_nveto_pmts, or a number, '
             'or a string like "pmt_commissioning_initial" which means calling '
             'hitfinder_thresholds.py, '
             'or a tuple like (correction=str, version=str, nT=boolean), '
             'which means we are using cmt.')

    min_split_nv = straxen.URLConfig(
        default=0.063, track=True, infer_type=False,
        help='Minimum height difference pe/sample between local minimum and maximum, '
             'that a pulse get split.')

    min_split_ratio_nv = straxen.URLConfig(
        default=0.75, track=True, infer_type=False,
        help='Min ratio between local maximum and minimum to split pulse (zero to switch this '
             'off).')

    entropy_template_nv = straxen.URLConfig(
        default='flat', track=True, infer_type=False,
        help='Template data is compared with in conditional entropy. Can be either "flat" or an '
             'template array.')

    entropy_square_data_nv = straxen.URLConfig(
        default=False, track=True, infer_type=False,
        help='Parameter which decides if data is first squared before normalized and compared to '
             'the template.')

    channel_map = straxen.URLConfig(track=False, type=immutabledict,
                                    help="immutabledict mapping subdetector to (min, max) "
                                         "channel number.")

    gain_model_nv = straxen.URLConfig(
        default="cmt://to_pe_model_nv?version=ONLINE&run_id=plugin.run_id", infer_type=False,
        help='PMT gain model. Specify as (model_type, model_config, nT = True)')

    def setup(self):
        self.channel_range = self.channel_map['nveto']
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1

        to_pe = self.gain_model_nv

        # Create to_pe array of size max channel:
        self.to_pe = np.zeros(self.channel_range[1] + 1, dtype=np.float32)
        self.to_pe[self.channel_range[0]:] = to_pe[:]

        # Check config of `hit_min_amplitude_nv` and define hit thresholds
        # if cmt config
        self.hit_thresholds = self.hit_min_amplitude_nv

    def compute(self, records_nv, start, end):
        records_nv = remove_switched_off_channels(records_nv, self.to_pe)
        hits = strax.find_hits(records_nv, min_amplitude=self.hit_thresholds)

        temp_hitlets = strax.create_hitlets_from_hits(hits,
                                                      self.save_outside_hits_nv,
                                                      self.channel_range,
                                                      chunk_start=start,
                                                      chunk_end=end)
        del hits

        # Get hitlet data and split hitlets:
        temp_hitlets = strax.get_hitlets_data(temp_hitlets,
                                              records_nv,
                                              to_pe=self.to_pe,
                                              min_hitlet_sample=600)

        temp_hitlets = strax.split_peaks(temp_hitlets,
                                         None,  # Only needed for peak splitting
                                         records_nv,
                                         None,  # Only needed for peak splitting
                                         self.to_pe,
                                         data_type='hitlets',
                                         algorithm='local_minimum',
                                         min_height=self.min_split_nv,
                                         min_ratio=self.min_split_ratio_nv,
                                         )

        # Compute other hitlet properties:
        # We have to loop here 3 times over all hitlets...
        strax.hitlet_properties(temp_hitlets)
        entropy = strax.conditional_entropy(temp_hitlets, template='flat', square_data=False)
        temp_hitlets['entropy'][:] = entropy

        # Remove data field:
        hitlets = np.zeros(len(temp_hitlets), dtype=strax.hitlet_dtype())
        strax.copy_to_buffer(temp_hitlets, hitlets, '_copy_hitlets')
        return strax.sort_by_time(hitlets)


def remove_switched_off_channels(hits, to_pe):
    """Removes hits which were found in a channel without any gain.
    :param hits: Hits found in records.
    :param to_pe: conversion factor from ADC per sample.
    :return: Hits
    """
    channel_off = np.argwhere(to_pe == 0).flatten()
    mask_off = np.isin(hits['channel'], channel_off)
    hits = hits[~mask_off]
    return hits
